import os
import sys

import click

from pyinfra import logger


def init_virtualenv():
    """
    Add a virtualenv to sys.path so the user can import modules from it.
    This isn't perfect: it doesn't use the Python interpreter with which the
    virtualenv was built, and it ignores the --no-site-packages option. A
    warning will appear suggesting the user installs IPython in the
    virtualenv, but for many cases, it probably works well enough.

    Adapted from IPython's implementation:
    https://github.com/ipython/ipython/blob/master/IPython/core/interactiveshell.py
    """

    if "VIRTUAL_ENV" not in os.environ:
        # Not in a virtualenv
        return

    p = os.path.normcase(sys.executable)
    p_venv = os.path.normcase(os.environ["VIRTUAL_ENV"])

    # executable path should end like /bin/python or \\scripts\\python.exe
    p_exe_up2 = os.path.dirname(os.path.dirname(p))
    if p_exe_up2 and os.path.samefile(p_exe_up2, p_venv):
        # Our exe is inside the virtualenv, don't need to do anything.
        return

    # fallback venv detection:
    # stdlib venv may symlink sys.executable, so we can't use realpath.
    # but others can symlink *to* the venv Python, so we can't just use sys.executable.
    # So we just check every item in the symlink tree (generally <= 3)
    paths = [p]
    while os.path.islink(p):
        p = os.path.normcase(os.path.join(os.path.dirname(p), os.readlink(p)))
        paths.append(p)

    # In Cygwin paths like 'c:\...' and '\cygdrive\c\...' are possible
    if p_venv.startswith("\\cygdrive"):
        p_venv = p_venv[11:]
    elif len(p_venv) >= 2 and p_venv[1] == ":":
        p_venv = p_venv[2:]

    if any(p_venv in p for p in paths):
        # Running properly in the virtualenv, don't need to do anything
        return

    logger.warning(
        (
            "Attempting to work in a virtualenv.\n"
            "    If you encounter problems, please install pyinfra inside the virtualenv."
        ),
    )
    click.echo(err=True)

    if sys.platform == "win32":
        virtual_env = os.path.join(
            os.environ["VIRTUAL_ENV"],
            "Lib",
            "site-packages",
        )
    else:
        virtual_env = os.path.join(
            os.environ["VIRTUAL_ENV"],
            "lib",
            "python%d.%d" % sys.version_info[:2],
            "site-packages",
        )

    import site

    sys.path.insert(0, virtual_env)
    site.addsitedir(virtual_env)



import json

from pyinfra.api import FactBase


class DockerFactBase(FactBase):
    abstract = True

    requires_command = "docker"

    def process(self, output):
        output = "".join(output)
        return json.loads(output)


class DockerSystemInfo(DockerFactBase):
    """
    Returns ``docker system info`` output in JSON format.
    """

    command = 'docker system info --format="{{json .}}"'


# All Docker objects
#


class DockerContainers(DockerFactBase):
    """
    Returns ``docker inspect`` output for all Docker containers.
    """

    command = "docker container inspect `docker ps -qa`"


class DockerImages(DockerFactBase):
    """
    Returns ``docker inspect`` output for all Docker images.
    """

    command = "docker image inspect `docker images -q`"


class DockerNetworks(DockerFactBase):
    """
    Returns ``docker inspect`` output for all Docker networks.
    """

    command = "docker network inspect `docker network ls -q`"


# Single Docker objects
#


class DockerSingleMixin(DockerFactBase):
    def command(self, object_id):
        return "docker {0} inspect {1} 2>&- || true".format(
            self.docker_type,
            object_id,
        )


class DockerContainer(DockerSingleMixin):
    """
    Returns ``docker inspect`` output for a single Docker container.
    """

    docker_type = "container"


class DockerImage(DockerSingleMixin):
    """
    Returns ``docker inspect`` output for a single Docker image.
    """

    docker_type = "image"


class DockerNetwork(DockerSingleMixin):
    """
    Returns ``docker inspect`` output for a single Docker network.
    """

    docker_type = "network"



"""
Manage MySQL databases, users and privileges.

Requires the ``mysql`` CLI executable on the target host(s).

All operations in this module take four optional arguments:
    + ``mysql_user``: the username to connect to mysql to
    + ``mysql_password``: the password for the connecting user
    + ``mysql_host``: the hostname of the server to connect to
    + ``mysql_port``: the port of the server to connect to

See the example/mysql.py

"""

from pyinfra import host
from pyinfra.api import MaskString, OperationError, StringCommand, operation
from pyinfra.facts.mysql import (
    MysqlDatabases,
    MysqlUserGrants,
    MysqlUsers,
    make_execute_mysql_command,
    make_mysql_command,
)


@operation(is_idempotent=False)
def sql(
    sql,
    database=None,
    # Details for speaking to MySQL via `mysql` CLI
    mysql_user=None,
    mysql_password=None,
    mysql_host=None,
    mysql_port=None,
):
    """
    Execute arbitrary SQL against MySQL.

    + sql: SQL command(s) to execute
    + database: optional database to open the connection with
    + mysql_*: global module arguments, see above
    """

    yield make_execute_mysql_command(
        sql,
        database=database,
        user=mysql_user,
        password=mysql_password,
        host=mysql_host,
        port=mysql_port,
    )


@operation
def user(
    user,
    present=True,
    user_hostname="localhost",
    password=None,
    privileges=None,
    # MySQL REQUIRE SSL/TLS options
    require=None,  # SSL or X509
    require_cipher=False,
    require_issuer=False,
    require_subject=False,
    # MySQL WITH resource limit options
    max_connections=None,
    max_queries_per_hour=None,
    max_updates_per_hour=None,
    max_connections_per_hour=None,
    # Details for speaking to MySQL via `mysql` CLI via `mysql` CLI
    mysql_user=None,
    mysql_password=None,
    mysql_host=None,
    mysql_port=None,
):
    """
    Add/remove/update MySQL users.

    + user: the name of the user
    + present: whether the user should exist or not
    + user_hostname: the hostname of the user
    + password: the password of the user (if created)
    + privileges: the global privileges for this user
    + mysql_*: global module arguments, see above

    Hostname:
        this + ``name`` makes the username - so changing this will create a new
        user, rather than update users with the same ``name``.

    Password:
        will only be applied if the user does not exist - ie pyinfra cannot
        detect if the current password doesn't match the one provided, so won't
        attempt to change it.

    **Example:**

    .. code:: python

        mysql.user(
            name="Create the pyinfra@localhost MySQL user",
            user="pyinfra",
            password="somepassword",
        )

        # Create a user with resource limits
        mysql.user(
            name="Create the pyinfra@localhost MySQL user",
            user="pyinfra",
            max_connections=50,
            max_updates_per_hour=10,
        )

        # Create a user that requires SSL for connections
        mysql.user(
            name="Create the pyinfra@localhost MySQL user",
            user="pyinfra",
            password="somepassword",
            require="SSL",
        )

        # Create a user that requires a specific certificate
        mysql.user(
            name="Create the pyinfra@localhost MySQL user",
            user="pyinfra",
            password="somepassword",
            require="X509",
            require_issuer="/C=SE/ST=Stockholm...",
            require_cipher="EDH-RSA-DES-CBC3-SHA",
        )
    """

    if require and require not in ("SSL", "X509"):
        raise OperationError('Invalid `require` value, must be: "SSL" or "X509"')

    if require != "X509":
        if require_cipher:
            raise OperationError('Cannot set `require_cipher` if `require` is not "X509"')
        if require_issuer:
            raise OperationError('Cannot set `require_issuer` if `require` is not "X509"')
        if require_subject:
            raise OperationError('Cannot set `require_subject` if `require` is not "X509"')

    current_users = host.get_fact(
        MysqlUsers,
        mysql_user=mysql_user,
        mysql_password=mysql_password,
        mysql_host=mysql_host,
        mysql_port=mysql_port,
    )

    user_host = "{0}@{1}".format(user, user_hostname)
    is_present = user_host in current_users

    if not present:
        if is_present:
            yield make_execute_mysql_command(
                'DROP USER "{0}"@"{1}"'.format(user, user_hostname),
                user=mysql_user,
                password=mysql_password,
                host=mysql_host,
                port=mysql_port,
            )
            current_users.pop(user_host)
        else:
            host.noop("mysql user {0}@{1} does not exist".format(user, user_hostname))
        return

    new_or_updated_user_fact = {
        "ssl_type": "ANY" if require == "SSL" else require,
        "ssl_cipher": require_cipher,
        "x509_issuer": require_issuer,
        "x509_subject": require_subject,
        "max_user_connections": max_connections,
        "max_questions": max_queries_per_hour,
        "max_updates": max_updates_per_hour,
        "max_connections": max_connections_per_hour,
    }

    if present and not is_present:
        sql_bits = ['CREATE USER "{0}"@"{1}"'.format(user, user_hostname)]
        if password:
            sql_bits.append(MaskString('IDENTIFIED BY "{0}"'.format(password)))

        if require == "SSL":
            sql_bits.append("REQUIRE SSL")

        if require == "X509":
            sql_bits.append("REQUIRE")
            require_bits = []

            if require_cipher:
                require_bits.append('CIPHER "{0}"'.format(require_cipher))
            if require_issuer:
                require_bits.append('ISSUER "{0}"'.format(require_issuer))
            if require_subject:
                require_bits.append('SUBJECT "{0}"'.format(require_subject))

            if not require_bits:
                require_bits.append("X509")

            sql_bits.extend(require_bits)

        resource_bits = []
        if max_connections:
            resource_bits.append("MAX_USER_CONNECTIONS {0}".format(max_connections))
        if max_queries_per_hour:
            resource_bits.append("MAX_QUERIES_PER_HOUR {0}".format(max_queries_per_hour))
        if max_updates_per_hour:
            resource_bits.append("MAX_UPDATES_PER_HOUR {0}".format(max_updates_per_hour))
        if max_connections_per_hour:
            resource_bits.append("MAX_CONNECTIONS_PER_HOUR {0}".format(max_connections_per_hour))

        if resource_bits:
            sql_bits.append("WITH")
            sql_bits.append(" ".join(resource_bits))

        yield make_execute_mysql_command(
            StringCommand(*sql_bits),
            user=mysql_user,
            password=mysql_password,
            host=mysql_host,
            port=mysql_port,
        )

        current_users[user_host] = new_or_updated_user_fact

    if present and is_present:
        current_user = current_users.get(user_host)

        alter_bits = []

        if require == "SSL":
            if current_user["ssl_type"] != "ANY":
                alter_bits.append("REQUIRE SSL")

        if require == "X509":
            require_bits = []

            if require_cipher and current_user["ssl_cipher"] != require_cipher:
                require_bits.append('CIPHER "{0}"'.format(require_cipher))
            if require_issuer and current_user["x509_issuer"] != require_issuer:
                require_bits.append('ISSUER "{0}"'.format(require_issuer))
            if require_subject and current_user["x509_subject"] != require_subject:
                require_bits.append('SUBJECT "{0}"'.format(require_subject))

            if not require_bits:
                if current_user["ssl_type"] != "X509":
                    require_bits.append("X509")

            if require_bits:
                alter_bits.append("REQUIRE")
                alter_bits.extend(require_bits)

        resource_bits = []
        if max_connections and current_user["max_user_connections"] != max_connections:
            resource_bits.append("MAX_USER_CONNECTIONS {0}".format(max_connections))
        if max_queries_per_hour and current_user["max_questions"] != max_queries_per_hour:
            resource_bits.append("MAX_QUERIES_PER_HOUR {0}".format(max_queries_per_hour))
        if max_updates_per_hour and current_user["max_updates"] != max_updates_per_hour:
            resource_bits.append("MAX_UPDATES_PER_HOUR {0}".format(max_updates_per_hour))
        if max_connections_per_hour and current_user["max_connections"] != max_connections_per_hour:
            resource_bits.append("MAX_CONNECTIONS_PER_HOUR {0}".format(max_connections_per_hour))

        if resource_bits:
            alter_bits.append("WITH")
            alter_bits.append(" ".join(resource_bits))

        if alter_bits:
            sql_bits = ['ALTER USER "{0}"@"{1}"'.format(user, user_hostname)]
            sql_bits.extend(alter_bits)
            yield make_execute_mysql_command(
                StringCommand(*sql_bits),
                user=mysql_user,
                password=mysql_password,
                host=mysql_host,
                port=mysql_port,
            )
            current_user.update(new_or_updated_user_fact)
        else:
            host.noop("mysql user {0}@{1} exists".format(user, user_hostname))

    # If we're here either the user exists or we just created them; either way
    # now we can check any privileges are set.
    if privileges:
        yield from _privileges(
            user,
            privileges,
            user_hostname=user_hostname,
            mysql_user=mysql_user,
            mysql_password=mysql_password,
            mysql_host=mysql_host,
            mysql_port=mysql_port,
        )


@operation
def database(
    database,
    # Desired database settings
    present=True,
    collate=None,
    charset=None,
    user=None,
    user_hostname="localhost",
    user_privileges="ALL",
    # Details for speaking to MySQL via `mysql` CLI
    mysql_user=None,
    mysql_password=None,
    mysql_host=None,
    mysql_port=None,
):
    """
    Add/remove MySQL databases.

    + name: the name of the database
    + present: whether the database should exist or not
    + collate: the collate to use when creating the database
    + charset: the charset to use when creating the database
    + user: MySQL user to grant privileges on this database to
    + user_hostname: the hostname of the MySQL user to grant
    + user_privileges: privileges to grant to any specified user
    + mysql_*: global module arguments, see above

    Collate/charset:
        these will only be applied if the database does not exist - ie pyinfra
        will not attempt to alter the existing databases collate/character sets.

    **Example:**

    .. code:: python

        mysql.database(
            name="Create the pyinfra_stuff database",
            database="pyinfra_stuff",
            user="pyinfra",
            user_privileges=["SELECT", "INSERT"],
            charset="utf8",
        )
    """

    current_databases = host.get_fact(
        MysqlDatabases,
        mysql_user=mysql_user,
        mysql_password=mysql_password,
        mysql_host=mysql_host,
        mysql_port=mysql_port,
    )

    is_present = database in current_databases

    if not present:
        if is_present:
            yield make_execute_mysql_command(
                "DROP DATABASE {0}".format(database),
                user=mysql_user,
                password=mysql_password,
                host=mysql_host,
                port=mysql_port,
            )
            current_databases.pop(database)
        else:
            host.noop("mysql database {0} does not exist".format(database))
        return

    # We want the database but it doesn't exist
    if present and not is_present:
        sql_bits = ["CREATE DATABASE {0}".format(database)]

        if collate:
            sql_bits.append("COLLATE {0}".format(collate))

        if charset:
            sql_bits.append("CHARSET {0}".format(charset))

        yield make_execute_mysql_command(
            " ".join(sql_bits),
            user=mysql_user,
            password=mysql_password,
            host=mysql_host,
            port=mysql_port,
        )
        current_databases[database] = {
            "collate": collate,
            "charset": charset,
        }
    else:
        host.noop("mysql database {0} exists".format(database))

    # Ensure any user privileges for this database
    if user and user_privileges:
        yield from privileges(
            user,
            user_hostname=user_hostname,
            privileges=user_privileges,
            database=database,
            mysql_user=mysql_user,
            mysql_password=mysql_password,
            mysql_host=mysql_host,
            mysql_port=mysql_port,
        )


@operation
def privileges(
    user,
    privileges,
    user_hostname="localhost",
    database="*",
    table="*",
    flush=True,
    with_grant_option=False,
    # Details for speaking to MySQL via `mysql` CLI
    mysql_user=None,
    mysql_password=None,
    mysql_host=None,
    mysql_port=None,
):
    """
    Add/remove MySQL privileges for a user, either global, database or table specific.

    + user: name of the user to manage privileges for
    + privileges: list of privileges the user should have (see also: ``with_grant_option`` argument)
    + user_hostname: the hostname of the user
    + database: name of the database to grant privileges to (defaults to all)
    + table: name of the table to grant privileges to (defaults to all)
    + flush: whether to flush (and update) the privileges table after any changes
    + with_grant_option: whether the grant option privilege should be set
    + mysql_*: global module arguments, see above
    """

    # Ensure we have a list
    if isinstance(privileges, str):
        privileges = {privileges}

    if isinstance(privileges, list):
        privileges = set(privileges)

    if with_grant_option:
        privileges.add("GRANT OPTION")

    if database != "*":
        database = "`{0}`".format(database)

    if table != "*":
        table = "`{0}`".format(table)

        # We can't set privileges on *.tablename as MySQL won't allow it
        if database == "*":
            raise OperationError(
                ("Cannot apply MySQL privileges on {0}.{1}, no database provided").format(
                    database,
                    table,
                ),
            )

    database_table = "{0}.{1}".format(database, table)
    user_grants = host.get_fact(
        MysqlUserGrants,
        user=user,
        hostname=user_hostname,
        mysql_user=mysql_user,
        mysql_password=mysql_password,
        mysql_host=mysql_host,
        mysql_port=mysql_port,
    )

    existing_privileges = set()
    if database_table in user_grants:
        existing_privileges = {
            "ALL" if privilege == "ALL PRIVILEGES" else privilege
            for privilege in user_grants[database_table]
        }
    else:
        user_grants[database_table] = set()

    def handle_privileges(action, target, privileges_to_apply):
        command = (
            "{action} {privileges} " "ON {database}.{table} " '{target} "{user}"@"{user_hostname}"'
        ).format(
            privileges=", ".join(sorted(privileges_to_apply)),
            action=action,
            target=target,
            database=database,
            table=table,
            user=user,
            user_hostname=user_hostname,
        )

        yield make_execute_mysql_command(
            command,
            user=mysql_user,
            password=mysql_password,
            host=mysql_host,
            port=mysql_port,
        )

    # Find / revoke any privileges that exist that do not match the desired state
    privileges_to_revoke = existing_privileges.difference(privileges)
    # Find / grant any privileges that we want but do not exist
    privileges_to_grant = privileges - existing_privileges

    # No privilege and we want it
    if privileges_to_grant:
        yield from handle_privileges("GRANT", "TO", privileges_to_grant)
        user_grants[database_table].update(privileges_to_grant)

    if privileges_to_revoke:
        yield from handle_privileges("REVOKE", "FROM", privileges_to_revoke)
        user_grants[database_table] -= privileges_to_revoke

    if privileges_to_grant or privileges_to_revoke:
        if flush:
            yield make_execute_mysql_command(
                "FLUSH PRIVILEGES",
                user=mysql_user,
                password=mysql_password,
                host=mysql_host,
                port=mysql_port,
            )
    else:
        host.noop("mysql privileges are already correct")


_privileges = privileges  # noqa: E305 (for use where kwarg is the same)


@operation(is_idempotent=False)
def dump(
    dest,
    database=None,
    # Details for speaking to MySQL via `mysql` CLI
    mysql_user=None,
    mysql_password=None,
    mysql_host=None,
    mysql_port=None,
):
    """
    Dump a MySQL database into a ``.sql`` file. Requires ``mysqldump``.

    + dest: name of the file to dump the SQL to
    + database: name of the database to dump
    + mysql_*: global module arguments, see above

    **Example:**

    .. code:: python

        mysql.dump(
            name="Dump the pyinfra_stuff database",
            dest="/tmp/pyinfra_stuff.dump",
            database="pyinfra_stuff",
        )
    """

    yield "{0} > {1}".format(
        make_mysql_command(
            executable="mysqldump",
            database=database,
            user=mysql_user,
            password=mysql_password,
            host=mysql_host,
            port=mysql_port,
        ),
        dest,
    )


@operation(is_idempotent=False)
def load(
    src,
    database=None,
    # Details for speaking to MySQL via `mysql` CLI
    mysql_user=None,
    mysql_password=None,
    mysql_host=None,
    mysql_port=None,
):
    """
    Load ``.sql`` file into a database.

    + src: the filename to read from
    + database: name of the database to import into
    + mysql_*: global module arguments, see above

    **Example:**

    .. code:: python

        mysql.load(
            name="Import the pyinfra_stuff dump into pyinfra_stuff_copy",
            src="/tmp/pyinfra_stuff.dump",
            database="pyinfra_stuff_copy",
        )
    """

    yield "{0} < {1}".format(
        make_mysql_command(
            database=database,
            user=mysql_user,
            password=mysql_password,
            host=mysql_host,
            port=mysql_port,
        ),
        src,
    )



import logging
import sys
import warnings
from fnmatch import fnmatch
from os import chdir as os_chdir, getcwd, path
from typing import Iterable, List, Tuple, Union

import click

from pyinfra import __version__, logger, state
from pyinfra.api import Config, State
from pyinfra.api.connect import connect_all, disconnect_all
from pyinfra.api.exceptions import NoGroupError, PyinfraError
from pyinfra.api.facts import get_facts
from pyinfra.api.operation import add_op
from pyinfra.api.operations import run_ops
from pyinfra.api.util import get_kwargs_str
from pyinfra.context import ctx_config, ctx_inventory, ctx_state
from pyinfra.operations import server

from .exceptions import CliError, UnexpectedExternalError, UnexpectedInternalError
from .inventory import make_inventory
from .log import setup_logging
from .prints import (
    print_facts,
    print_inventory,
    print_meta,
    print_results,
    print_state_facts,
    print_state_operations,
    print_support_info,
)
from .util import (
    exec_file,
    get_facts_and_args,
    get_operation_and_args,
    load_deploy_file,
    parse_cli_arg,
)
from .virtualenv import init_virtualenv


def _exit():
    if ctx_state.isset() and state.failed_hosts:
        sys.exit(1)
    sys.exit(0)


def _print_support(ctx, param, value):
    if not value:
        return

    click.echo("--> Support information:", err=True)
    print_support_info()
    ctx.exit()


@click.command()
@click.argument("inventory", nargs=1, type=click.Path(exists=False))
@click.argument("operations", nargs=-1, required=True, type=click.Path(exists=False))
@click.option(
    "verbosity",
    "-v",
    count=True,
    help="Print meta (-v), input (-vv) and output (-vvv).",
)
@click.option(
    "--dry",
    is_flag=True,
    default=False,
    help="Don't execute operations on the target hosts.",
)
@click.option(
    "--limit",
    help="Restrict the target hosts by name and group name.",
    multiple=True,
)
@click.option("--fail-percent", type=int, help="% of hosts allowed to fail.")
@click.option(
    "--data",
    multiple=True,
    help="Override data values, format key=value.",
)
@click.option(
    "--group-data",
    multiple=True,
    help="Paths to load additional group data from (overrides matching keys).",
)
@click.option(
    "--config",
    "config_filename",
    help="Specify config file to use (default: config.py).",
    default="config.py",
)
@click.option(
    "--chdir",
    help="Set the working directory before executing.",
)
# Auth args
@click.option(
    "--sudo",
    is_flag=True,
    default=False,
    help="Whether to execute operations with sudo.",
)
@click.option("--sudo-user", help="Which user to sudo when sudoing.")
@click.option(
    "--use-sudo-password",
    is_flag=True,
    default=False,
    help="Whether to use a password with sudo.",
)
@click.option("--su-user", help="Which user to su to.")
@click.option("--shell-executable", help='Shell to use (ex: "sh", "cmd", "ps").')
# Operation flow args
@click.option("--parallel", type=int, help="Number of operations to run in parallel.")
@click.option(
    "--no-wait",
    is_flag=True,
    default=False,
    help="Don't wait between operations for hosts.",
)
@click.option(
    "--serial",
    is_flag=True,
    default=False,
    help="Run operations in serial, host by host.",
)
# SSH connector args
# TODO: remove the non-ssh-prefixed variants
@click.option("--ssh-user", "--user", "ssh_user", help="SSH user to connect as.")
@click.option("--ssh-port", "--port", "ssh_port", type=int, help="SSH port to connect to.")
@click.option("--ssh-key", "--key", "ssh_key", type=click.Path(), help="SSH Private key filename.")
@click.option(
    "--ssh-key-password",
    "--key-password",
    "ssh_key_password",
    help="SSH Private key password.",
)
@click.option("--ssh-password", "--password", "ssh_password", help="SSH password.")
# WinRM connector args
@click.option("--winrm-username", help="WINRM user to connect as.")
@click.option("--winrm-password", help="WINRM password.")
@click.option("--winrm-port", help="WINRM port to connect to.")
@click.option("--winrm-transport", help="WINRM transport for use.")
# Eager commands (pyinfra --support)
@click.option(
    "--support",
    is_flag=True,
    is_eager=True,
    callback=_print_support,
    help="Print useful information for support and exit.",
)
# Debug args
@click.option(
    "--quiet",
    is_flag=True,
    default=False,
    help="Hide most pyinfra output.",
)
@click.option(
    "--debug",
    is_flag=True,
    default=False,
    help="Print debug info.",
)
@click.option(
    "--debug-facts",
    is_flag=True,
    default=False,
    help="Print facts after generating operations and exit.",
)
@click.option(
    "--debug-operations",
    is_flag=True,
    default=False,
    help="Print operations after generating and exit.",
)
@click.version_option(
    version=__version__,
    prog_name="pyinfra",
    message="%(prog)s: v%(version)s",
)
def cli(*args, **kwargs):
    """
    pyinfra manages the state of one or more servers. It can be used for
    app/service deployment, config management and ad-hoc command execution.

    Documentation: pyinfra.readthedocs.io

    # INVENTORY

    \b
    + a file (inventory.py)
    + hostname (host.net)
    + Comma separated hostnames:
      host-1.net,host-2.net,@local

    # OPERATIONS

    \b
    # Run one or more deploys against the inventory
    pyinfra INVENTORY deploy_web.py [deploy_db.py]...

    \b
    # Run a single operation against the inventory
    pyinfra INVENTORY server.user pyinfra home=/home/pyinfra

    \b
    # Execute an arbitrary command against the inventory
    pyinfra INVENTORY exec -- echo "hello world"

    \b
    # Run one or more facts against the inventory
    pyinfra INVENTORY fact server.LinuxName [server.Users]...
    pyinfra INVENTORY fact files.File path=/path/to/file...

    \b
    # Debug the inventory hosts and data
    pyinfra INVENTORY debug-inventory
    """

    try:
        _main(*args, **kwargs)

    except PyinfraError as e:
        # Re-raise any internal exceptions that aren't handled by click as
        # CliErrors which are.
        if not isinstance(e, click.ClickException):
            message = getattr(e, "message", e.args[0])
            raise CliError(message)

        raise

    except UnexpectedExternalError:
        # Pass unexpected external exceptions through as-is
        raise

    except Exception as e:
        # Re-raise any unexpected internal exceptions as UnexpectedInternalError
        raise UnexpectedInternalError(e)

    finally:
        if ctx_state.isset() and state.initialised:
            # Triggers any executor disconnect requirements
            disconnect_all(state)


def _main(
    inventory,
    operations: Union[List, Tuple],
    verbosity: int,
    chdir: str,
    ssh_user,
    ssh_port: int,
    ssh_key,
    ssh_key_password: str,
    ssh_password: str,
    winrm_username: str,
    winrm_password: str,
    winrm_port,
    winrm_transport,
    shell_executable,
    sudo: bool,
    sudo_user: str,
    use_sudo_password: bool,
    su_user: str,
    parallel: int,
    fail_percent: int,
    data,
    group_data,
    config_filename: str,
    dry: bool,
    limit: Iterable,
    no_wait: bool,
    serial: bool,
    quiet: bool,
    debug: bool,
    debug_facts: bool,
    debug_operations: bool,
    support: bool = None,
):
    # Setup working directory
    #
    if chdir:
        os_chdir(chdir)

    # Setup logging & Bootstrap/Venv
    #
    _setup_log_level(debug, quiet)
    init_virtualenv()

    # Check operations are valid and setup commands
    #
    original_operations, operations, command, chdir = _validate_operations(operations, chdir)

    # Setup state, config & inventory
    #
    state = _setup_state(verbosity, quiet)
    config = Config()
    ctx_config.set(config)

    # Update Config & Override Data
    #
    config = _set_config(
        config,
        config_filename,
        sudo,
        sudo_user,
        use_sudo_password,
        su_user,
        parallel,
        shell_executable,
        fail_percent,
        quiet,
    )
    override_data = _set_override_data(
        data,
        ssh_user,
        ssh_key,
        ssh_key_password,
        ssh_port,
        ssh_password,
        winrm_username,
        winrm_password,
        winrm_port,
        winrm_transport,
    )

    # Load up the inventory from the filesystem
    #
    echo_msg("--> Loading inventory...", quiet)
    inventory, inventory_group = make_inventory(
        inventory,
        cwd=state.cwd,
        override_data=override_data,
        group_data_directories=group_data,
    )
    ctx_inventory.set(inventory)

    # Now that we have inventory, apply --limit config override
    initial_limit = _apply_inventory_limit(inventory, limit)

    # Initialise the state
    state.init(inventory, config, initial_limit=initial_limit)

    if command == "debug-inventory":
        print_inventory(state)
        _exit()

    # Connect to the hosts & start handling the user commands
    #
    echo_msg("--> Connecting to hosts...", quiet)
    connect_all(state)
    state, config = _handle_commands(state, config, command, original_operations, operations, quiet)

    # Print proposed changes, execute unless --dry, and exit
    #
    echo_msg("--> Proposed changes:", quiet)
    print_meta(state)

    # If --debug-facts or --debug-operations, print and exit
    if debug_facts or debug_operations:
        if debug_facts:
            print_state_facts(state)

        if debug_operations:
            print_state_operations(state)

        _exit()

    if dry:
        _exit()

    echo_msg(quiet=quiet)
    echo_msg("--> Beginning operation run...", quiet)

    run_ops(state, serial=serial, no_wait=no_wait)

    echo_msg("--> Results:", quiet)
    print_results(state)
    _exit()


# Setup
#
def _setup_log_level(debug, quiet):
    if not debug and not sys.warnoptions:
        warnings.simplefilter("ignore")

    log_level = logging.INFO
    if debug:
        log_level = logging.DEBUG
    elif quiet:
        log_level = logging.WARNING

    setup_logging(log_level)


def _validate_operations(operations, chdir):

    # Make a copy before we overwrite
    original_operations = operations

    # Debug (print) inventory + group data
    if operations[0] == "debug-inventory":
        command = "debug-inventory"

    # Get one or more facts
    elif operations[0] == "fact":
        command = "fact"
        operations = get_facts_and_args(operations[1:])

    # Execute a raw command with server.shell
    elif operations[0] == "exec":
        command = "exec"
        operations = operations[1:]

    # Execute one or more deploy files
    elif all(cmd.endswith(".py") for cmd in operations):
        command = "deploy"

        filenames = []

        for filename in operations[0:]:
            if path.exists(filename):
                filenames.append(filename)
                continue
            if chdir and filename.startswith(chdir):
                correct_filename = path.relpath(filename, chdir)
                logger.warning(
                    (
                        "Fixing deploy filename under `--chdir` argument: "
                        f"{filename} -> {correct_filename}"
                    ),
                )
                filenames.append(correct_filename)
                continue
            raise CliError(
                "No deploy file: {0}".format(
                    path.join(chdir, filename) if chdir else filename,
                ),
            )

        operations = filenames

    # Operation w/optional args (<module>.<op> ARG1 ARG2 ...)
    elif len(operations[0].split(".")) == 2:
        command = "op"
        operations = get_operation_and_args(operations)

    else:
        raise CliError(
            """Invalid operations: {0}

    Operation usage:
    pyinfra INVENTORY deploy_web.py [deploy_db.py]...
    pyinfra INVENTORY server.user pyinfra home=/home/pyinfra
    pyinfra INVENTORY exec -- echo "hello world"
    pyinfra INVENTORY fact os [users]...""".format(
                operations,
            ),
        )

    return original_operations, operations, command, chdir


def _set_verbosity(state, verbosity, quiet):
    if verbosity > 0:
        state.print_fact_info = True
        state.print_noop_info = True

    if verbosity > 1:
        state.print_input = state.print_fact_input = True

    if verbosity > 2:
        state.print_output = state.print_fact_output = True

    return state


def _setup_state(verbosity, quiet):
    cwd = getcwd()
    if cwd not in sys.path:  # ensure cwd is present in sys.path
        sys.path.append(cwd)

    state = State()
    state.cwd = cwd
    ctx_state.set(state)

    state = _set_verbosity(state, verbosity, quiet)
    return state


def _set_config(
    config,
    config_filename,
    sudo,
    sudo_user,
    use_sudo_password,
    su_user,
    parallel,
    shell_executable,
    fail_percent,
    quiet,
):
    echo_msg("--> Loading config...", quiet)

    # Load up any config.py from the filesystem
    config_filename = path.join(state.cwd, config_filename)
    if path.exists(config_filename):
        exec_file(config_filename)

    # Lock the current config, this allows us to restore this version after
    # executing deploy files that may alter them.
    config.lock_current_state()

    # Arg based config overrides
    if sudo:
        config.SUDO = True
        if sudo_user:
            config.SUDO_USER = sudo_user

    if use_sudo_password:
        config.USE_SUDO_PASSWORD = use_sudo_password

    if su_user:
        config.SU_USER = su_user

    if parallel:
        config.PARALLEL = parallel

    if shell_executable:
        config.SHELL = None if shell_executable in ("None", "null") else shell_executable

    if fail_percent is not None:
        config.FAIL_PERCENT = fail_percent

    return config


def _set_override_data(
    data,
    ssh_user,
    ssh_key,
    ssh_key_password,
    ssh_port,
    ssh_password,
    winrm_username,
    winrm_password,
    winrm_port,
    winrm_transport,
):
    override_data = {}

    for arg in data:
        key, value = arg.split("=", 1)
        override_data[key] = value

    override_data = {key: parse_cli_arg(value) for key, value in override_data.items()}

    for key, value in (
        ("ssh_user", ssh_user),
        ("ssh_key", ssh_key),
        ("ssh_key_password", ssh_key_password),
        ("ssh_port", ssh_port),
        ("ssh_password", ssh_password),
        ("winrm_username", winrm_username),
        ("winrm_password", winrm_password),
        ("winrm_port", winrm_port),
        ("winrm_transport", winrm_transport),
    ):
        if value:
            override_data[key] = value

    return override_data


def _apply_inventory_limit(inventory, limit):
    initial_limit = None
    if limit:
        all_limit_hosts = []

        for limiter in limit:
            try:
                limit_hosts = inventory.get_group(limiter)
            except NoGroupError:
                limit_hosts = [host for host in inventory if fnmatch(host.name, limiter)]

            if not limit_hosts:
                logger.warning("No host matches found for --limit pattern: {0}".format(limiter))

            all_limit_hosts.extend(limit_hosts)
        initial_limit = list(set(all_limit_hosts))

    return initial_limit


# Operations Execution
#
def _handle_commands(state, config, command, original_operations, operations, quiet):

    if command == "fact":
        state, fact_data = _run_fact_operations(state, config, operations, quiet)
        print_facts(fact_data)
        _exit()

    if command == "exec":
        state = _run_exec_operations(state, config, operations, quiet)

    elif command == "deploy":
        state, config, operations = _run_deploy_operations(state, config, operations, quiet)

    elif command == "op":
        state, kwargs = _run_op_operations(state, config, operations, original_operations, quiet)

    return state, config


def _run_fact_operations(state, config, operations, quiet):
    echo_msg("--> Gathering facts...", quiet)

    state.print_fact_info = True
    fact_data = {}

    for i, command in enumerate(operations):
        fact_cls, args, kwargs = command
        fact_key = fact_cls.name

        if args or kwargs:
            _fact_args = args or ""
            _fact_details = " ({0})".format(get_kwargs_str(kwargs)) if kwargs else ""
            fact_key = "{0}{1}{2}".format(fact_cls.name, _fact_args, _fact_details)

        try:
            fact_data[fact_key] = get_facts(
                state,
                fact_cls,
                args=args,
                kwargs=kwargs,
                apply_failed_hosts=False,
            )
        except PyinfraError:
            pass

    return state, fact_data


def _run_exec_operations(state, config, operations, quiet):
    state.print_output = True
    add_op(
        state,
        server.shell,
        " ".join(operations),
        _allow_cli_mode=True,
    )
    return state


def _run_deploy_operations(state, config, operations, quiet):
    echo_msg("--> Preparing Operations...", quiet)

    # Number of "steps" to make = number of files * number of hosts
    for i, filename in enumerate(operations):
        _log_styled_msg = click.style(filename, bold=True)
        logger.info("Loading: {0}".format(_log_styled_msg))

        state.current_op_file_number = i
        load_deploy_file(state, filename)

        # Remove any config changes introduced by the deploy file & any includes
        config.reset_locked_state()

    return state, config, operations


def _run_op_operations(state, config, operations, original_operations, quiet):
    echo_msg("--> Preparing operation...", quiet)

    op, args = operations
    args, kwargs = args
    kwargs["_allow_cli_mode"] = True

    def print_host_ready(host):
        _ready_style = click.style("Ready:", "green")
        _original_op_msg = click.style(original_operations[0], bold=True)
        logger.info("{0}{1} {2}".format(host.print_prefix, _ready_style, _original_op_msg))

    kwargs["_after_host_callback"] = print_host_ready

    add_op(state, op, *args, **kwargs)

    return state, kwargs


# Utils
#
def echo_msg(msg=None, quiet=None):
    if not quiet:
        click.echo(err=True)
    if msg:
        click.echo(msg, err=True)



"""
Manage yum packages and repositories. Note that yum package names are case-sensitive.
"""

from pyinfra import host, state
from pyinfra.api import operation
from pyinfra.facts.rpm import RpmPackageProvides, RpmPackages

from . import files
from .util.packaging import ensure_packages, ensure_rpm, ensure_yum_repo


@operation(is_idempotent=False)
def key(src):
    """
    Add yum gpg keys with ``rpm``.

    + src: filename or URL

    Note:
        always returns one command, not state checking

    **Example:**

    .. code:: python

        linux_id = host.get_fact(LinuxDistribution)["release_meta"].get("ID")
        yum.key(
            name="Add the Docker CentOS gpg key",
            src=f"https://download.docker.com/linux/{linux_id}/gpg",
        )

    """

    yield "rpm --import {0}".format(src)


@operation
def repo(
    src,
    present=True,
    baseurl=None,
    description=None,
    enabled=True,
    gpgcheck=True,
    gpgkey=None,
):
    # NOTE: if updating this docstring also update `dnf.repo`
    """
    Add/remove/update yum repositories.

    + src: URL or name for the ``.repo``   file
    + present: whether the ``.repo`` file should be present
    + baseurl: the baseurl of the repo (if ``name`` is not a URL)
    + description: optional verbose description
    + enabled: whether this repo is enabled
    + gpgcheck: whether set ``gpgcheck=1``
    + gpgkey: the URL to the gpg key for this repo

    ``Baseurl``/``description``/``gpgcheck``/``gpgkey``:
        These are only valid when ``src`` is a filename (ie not a URL). This is
        for manual construction of repository files. Use a URL to download and
        install remote repository files.

    **Examples:**

    .. code:: python

        # Download a repository file
        yum.repo(
            name="Install Docker-CE repo via URL",
            src="https://download.docker.com/linux/centos/docker-ce.repo",
        )

        # Create the repository file from baseurl/etc
        yum.repo(
            name="Add the Docker CentOS repo",
            src="DockerCE",
            baseurl="https://download.docker.com/linux/centos/7/$basearch/stable",
        )
    """

    yield from ensure_yum_repo(
        state,
        host,
        files,
        src,
        baseurl,
        present,
        description,
        enabled,
        gpgcheck,
        gpgkey,
    )


@operation
def rpm(src, present=True):
    # NOTE: if updating this docstring also update `dnf.rpm`
    """
    Add/remove ``.rpm`` file packages.

    + src: filename or URL of the ``.rpm`` package
    + present: whether ore not the package should exist on the system

    URL sources with ``present=False``:
        If the ``.rpm`` file isn't downloaded, pyinfra can't remove any existing
        package as the file won't exist until mid-deploy.

    **Example:**

    .. code:: python

        major_version = host.get_fact(LinuxDistribution)["major"]
        dnf.rpm(
           name="Install EPEL rpm to enable EPEL repo",
           src=f"https://dl.fedoraproject.org/pub/epel/epel-release-latest-{major_version}.noarch.rpm",
        )
    """

    yield from ensure_rpm(state, host, files, src, present, "yum")


@operation(is_idempotent=False)
def update():
    """
    Updates all yum packages.
    """

    yield "yum update -y"


_update = update  # noqa: E305 (for use below where update is a kwarg)


@operation
def packages(
    packages=None,
    present=True,
    latest=False,
    update=False,
    clean=False,
    nobest=False,
    extra_install_args=None,
    extra_uninstall_args=None,
):
    """
    Install/remove/update yum packages & updates.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + update: run ``yum update`` before installing packages
    + clean: run ``yum clean all`` before installing packages
    + nobest: add the no best option to install
    + extra_install_args: additional arguments to the yum install command
    + extra_uninstall_args: additional arguments to the yum uninstall command

    Versions:
        Package versions can be pinned as follows: ``<pkg>=<version>``

    **Examples:**

    .. code:: python

        # Update package list and install packages
        yum.packages(
            name="Install Vim and Vim enhanced",
            packages=["vim-enhanced", "vim"],
            update=True,
        )

        # Install the latest versions of packages (always check)
        yum.packages(
            name="Install latest Vim",
            packages=["vim"],
            latest=True,
        )
    """

    if clean:
        yield "yum clean all"

    if update:
        yield from _update()

    install_command = ["yum", "install", "-y"]

    if nobest:
        install_command.append("--nobest")

    if extra_install_args:
        install_command.append(extra_install_args)

    uninstall_command = ["yum", "remove", "-y"]

    if extra_uninstall_args:
        uninstall_command.append(extra_uninstall_args)

    yield from ensure_packages(
        host,
        packages,
        host.get_fact(RpmPackages),
        present,
        install_command=" ".join(install_command),
        uninstall_command=" ".join(uninstall_command),
        upgrade_command="yum update -y",
        version_join="=",
        latest=latest,
        expand_package_fact=lambda package: host.get_fact(
            RpmPackageProvides,
            name=package,
        ),
    )



from pyinfra.api import FactBase


class LaunchdStatus(FactBase):
    """
    Returns a dict of name -> status for launchd managed services.
    """

    command = "launchctl list"
    requires_command = "launchctl"

    default = dict

    def process(self, output):
        services = {}

        for line in output:
            bits = line.split()

            if not bits or bits[0] == "PID":
                continue

            name = bits[2]
            status = False

            try:
                int(bits[0])
                status = True
            except ValueError:
                pass

            services[name] = status

        return services



"""
Execute commands and up/download files *from* the remote host.

Eg: ``pyinfra -> inventory-host.net <-> another-host.net``
"""

import shlex

from pyinfra import host, state
from pyinfra.api import OperationError, operation
from pyinfra.facts.files import File, FindInFile
from pyinfra.facts.server import Home

from . import files


@operation
def keyscan(hostname, force=False, port=22):
    """
    Check/add hosts to the ``~/.ssh/known_hosts`` file.

    + hostname: hostname that should have a key in ``known_hosts``
    + force: if the key already exists, remove and rescan

    **Example:**

    .. code:: python

        ssh.keyscan(
            name="Set add server two to known_hosts on one",
            hostname="two.example.com",
        )
    """

    homedir = host.get_fact(Home)

    yield from files.directory(
        "{0}/.ssh".format(homedir),
        mode=700,
    )

    hostname_present = host.get_fact(
        FindInFile,
        path="{0}/.ssh/known_hosts".format(homedir),
        pattern=hostname,
    )

    did_keyscan = False
    keyscan_command = "ssh-keyscan -p {0} {1} >> {2}/.ssh/known_hosts".format(
        port,
        hostname,
        homedir,
    )

    if not hostname_present:
        yield keyscan_command
        did_keyscan = True

    elif force:
        yield "ssh-keygen -R {0}".format(hostname)
        yield keyscan_command
        did_keyscan = True

    else:
        host.noop("host key for {0} already exists".format(hostname))

    if did_keyscan:
        host.create_fact(
            FindInFile,
            kwargs={"path": "{0}/.ssh/known_hosts".format(homedir), "pattern": hostname},
            data=["{0} unknown unknown".format(hostname)],
        )


@operation(is_idempotent=False)
def command(hostname, command, user=None, port=22):
    """
    Execute commands on other servers over SSH.

    + hostname: the hostname to connect to
    + command: the command to execute
    + user: connect with this user
    + port: connect to this port

    **Example:**

    .. code:: python

        ssh.command(
            name="Create file by running echo from host one to host two",
            hostname="two.example.com",
            command="echo "one was here" > /tmp/one.txt",
            user="vagrant",
        )
    """

    command = shlex.quote(command)

    connection_target = hostname
    if user:
        connection_target = "@".join((user, hostname))

    yield "ssh -p {0} {1} {2}".format(port, connection_target, command)


@operation(is_idempotent=False)
def upload(
    hostname,
    filename,
    remote_filename=None,
    port=22,
    user=None,
    use_remote_sudo=False,
    ssh_keyscan=False,
):
    """
    Upload files to other servers using ``scp``.

    + hostname: hostname to upload to
    + filename: file to upload
    + remote_filename: where to upload the file to (defaults to ``filename``)
    + port: connect to this port
    + user: connect with this user
    + use_remote_sudo: upload to a temporary location and move using sudo
    + ssh_keyscan: execute ``ssh.keyscan`` before uploading the file
    """

    remote_filename = remote_filename or filename

    # Figure out where we're connecting (host or user@host)
    connection_target = hostname
    if user:
        connection_target = "@".join((user, hostname))

    if ssh_keyscan:
        yield from keyscan(hostname)

    # If we're not using sudo on the remote side, just scp the file over
    if not use_remote_sudo:
        yield "scp -P {0} {1} {2}:{3}".format(
            port,
            filename,
            connection_target,
            remote_filename,
        )

    else:
        # Otherwise - we need a temporary location for the file
        temp_remote_filename = state.get_temp_filename()

        # scp it to the temporary location
        upload_cmd = "scp -P {0} {1} {2}:{3}".format(
            port,
            filename,
            connection_target,
            temp_remote_filename,
        )

        yield upload_cmd

        # And sudo sudo to move it
        yield from command(
            hostname=hostname,
            command="sudo mv {0} {1}".format(temp_remote_filename, remote_filename),
            port=port,
            user=user,
        )


@operation
def download(
    hostname,
    filename,
    local_filename=None,
    force=False,
    port=22,
    user=None,
    ssh_keyscan=False,
):
    """
    Download files from other servers using ``scp``.

    + hostname: hostname to upload to
    + filename: file to download
    + local_filename: where to download the file to (defaults to ``filename``)
    + force: always download the file, even if present locally
    + port: connect to this port
    + user: connect with this user
    + ssh_keyscan: execute ``ssh.keyscan`` before uploading the file
    """

    local_filename = local_filename or filename

    # Get local file info
    local_file_info = host.get_fact(File, path=local_filename)

    # Local file exists but isn't a file?
    if local_file_info is False:
        raise OperationError(
            "Local destination {0} already exists and is not a file".format(
                local_filename,
            ),
        )

    # If the local file exists and we're not forcing a re-download, no-op
    if local_file_info and not force:
        host.noop("file {0} is already downloaded".format(filename))
        return

    # Figure out where we're connecting (host or user@host)
    connection_target = hostname
    if user:
        connection_target = "@".join((user, hostname))

    if ssh_keyscan:
        yield from keyscan(hostname)

    # Download the file with scp
    yield "scp -P {0} {1}:{2} {3}".format(
        port,
        connection_target,
        filename,
        local_filename,
    )
    host.create_fact(
        File,
        kwargs={"path": local_filename},
        data={"mode": None, "group": None, "user": user, "mtime": None},
    )



# encoding: utf8

from pyinfra.api import FactBase

from .util.packaging import parse_packages

NPM_REGEX = r"^[└├]\─\─\s([a-zA-Z0-9\-]+)@([0-9\.]+)$"


class NpmPackages(FactBase):
    """
    Returns a dict of installed npm packages globally or in a given directory:

    .. code:: python

        {
            "package_name": ["version"],
        }
    """

    default = dict

    requires_command = "npm"

    def command(self, directory=None):
        if directory:
            return ("cd {0} && npm list -g --depth=0").format(directory)
        else:
            return "npm list -g --depth=0"

    def process(self, output):
        return parse_packages(NPM_REGEX, output)



from pyinfra.api import FactBase

from .util import make_cat_files_command
from .util.packaging import parse_yum_repositories


class DnfRepositories(FactBase):
    """
    Returns a list of installed dnf repositories:

    .. code:: python

        [
            {
                "name": "CentOS-$releasever - Apps",
                "baseurl": "http://mirror.centos.org/$contentdir/$releasever/Apps/$basearch/os/",
                "gpgcheck": "1",
                "enabled": "1",
                "gpgkey": "file:///etc/pki/rpm-gpg/RPM-GPG-KEY-centosofficial",
            },
        ]
    """

    command = make_cat_files_command(
        "/etc/dnf.conf",
        "/etc/dnf.repos.d/*.repo",
        "/etc/yum.repos.d/*.repo",
    )
    requires_command = "dnf"

    default = list

    def process(self, output):
        return parse_yum_repositories(output)



from contextlib import contextmanager
from graphlib import CycleError, TopologicalSorter
from multiprocessing import cpu_count
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set
from uuid import uuid4

from gevent.pool import Pool
from paramiko import PKey

if TYPE_CHECKING:
    from pyinfra.api.host import Host
    from pyinfra.api.inventory import Inventory

from pyinfra import logger

from .config import Config
from .exceptions import PyinfraError
from .util import sha1_hash

# Work out the max parallel we can achieve with the open files limit of the user/process,
# take 10 for opening Python files and /3 for ~3 files per host during op runs.
# See: https://github.com/Fizzadar/pyinfra/issues/44
try:
    from resource import RLIMIT_NOFILE, getrlimit

    nofile_limit, _ = getrlimit(RLIMIT_NOFILE)
    MAX_PARALLEL = round((nofile_limit - 10) / 3)

# Resource isn't available on Windows
except ImportError:
    nofile_limit = 0
    MAX_PARALLEL = 100000


class BaseStateCallback(object):
    # Host callbacks
    #

    @staticmethod
    def host_before_connect(state: "State", host: "Host"):
        pass

    @staticmethod
    def host_connect(state: "State", host: "Host"):
        pass

    @staticmethod
    def host_connect_error(state: "State", host: "Host", error):
        pass

    @staticmethod
    def host_disconnect(state: "State", host: "Host"):
        pass

    # Operation callbacks
    #

    @staticmethod
    def operation_start(state: "State", op_hash):
        pass

    @staticmethod
    def operation_host_start(state: "State", host: "Host", op_hash):
        pass

    @staticmethod
    def operation_host_success(state: "State", host: "Host", op_hash):
        pass

    @staticmethod
    def operation_host_error(state: "State", host: "Host", op_hash):
        pass

    @staticmethod
    def operation_end(state: "State", op_hash):
        pass


class State(object):
    """
    Manages state for a pyinfra deploy.
    """

    initialised: bool = False

    # A pyinfra.api.Inventory which stores all our pyinfra.api.Host's
    inventory: "Inventory"

    # A pyinfra.api.Config
    config: "Config"

    # Main gevent pool
    pool: "Pool"

    # Whether we are executing operations (ie hosts are all ready)
    is_executing: bool = False

    print_noop_info: bool = False  # print "[host] noop: reason for noop"
    print_fact_info: bool = False  # print "loaded fact X"
    print_input: bool = False
    print_fact_input: bool = False
    print_output: bool = False
    print_fact_output: bool = False

    # Used in CLI
    cwd: Optional[str] = None  # base directory for locating files/templates/etc
    current_deploy_filename: Optional[str] = None
    current_exec_filename: Optional[str] = None
    current_op_file_number: int = 0

    def __init__(
        self, inventory: Optional["Inventory"] = None, config: Optional["Config"] = None, **kwargs
    ):
        """Initializes the state, the main Pyinfra

        Args:
            inventory (Optional[Inventory], optional): The inventory. Defaults to None.
            config (Optional[Config], optional): The config object. Defaults to None.
        """
        if inventory:
            self.init(inventory, config, **kwargs)

    def init(self, inventory: "Inventory", config: Optional["Config"], initial_limit=None):
        # Config validation
        #

        # If no config, create one using the defaults
        if config is None:
            config = Config()

        if not config.PARALLEL:
            # TODO: benchmark this
            # In my own tests the optimum number of parallel SSH processes is
            # ~20 per CPU core - no science here yet, needs benchmarking!
            cpus = cpu_count()
            ideal_parallel = cpus * 20

            config.PARALLEL = min(ideal_parallel, len(inventory), MAX_PARALLEL)

        # If explicitly set, just issue a warning
        elif config.PARALLEL > MAX_PARALLEL:
            logger.warning(
                (
                    "Parallel set to {0}, but this may hit the open files limit of {1}.\n"
                    "    Max recommended value: {2}"
                ).format(config.PARALLEL, nofile_limit, MAX_PARALLEL),
            )

        # Actually initialise the state object
        #

        self.callback_handlers: List[BaseStateCallback] = []

        # Setup greenlet pools
        self.pool = Pool(config.PARALLEL)
        self.fact_pool = Pool(config.PARALLEL)

        # Private keys
        self.private_keys: Dict[str, PKey] = {}

        # Assign inventory/config
        self.inventory = inventory
        self.config = config

        # Hosts we've activated at any time
        self.activated_hosts: Set["Host"] = set()
        # Active hosts that *haven't* failed yet
        self.active_hosts: Set["Host"] = set()
        # Hosts that have failed
        self.failed_hosts: Set["Host"] = set()

        # Limit hosts changes dynamically to limit operations to a subset of hosts
        self.limit_hosts: List["Host"] = initial_limit

        # Op basics
        self.op_meta: Dict[str, dict] = {}  # maps operation hash -> names/etc
        self.ops_run: Set[str] = set()  # list of ops which have been started/run

        # Op dict for each host
        self.ops: Dict["Host", dict] = {host: {} for host in inventory}

        # Facts dict for each host
        self.facts: Dict["Host", Any] = {host: {} for host in inventory}

        # Meta dict for each host
        self.meta = {
            host: {
                "ops": 0,  # one function call in a deploy file
                "ops_change": 0,
                "ops_no_change": 0,
                "commands": 0,  # actual # of commands to run
                "op_hashes": set(),
            }
            for host in inventory
        }

        # Results dict for each host
        self.results = {
            host: {
                "ops": 0,  # success_ops + failed ops w/ignore_errors
                "success_ops": 0,
                "error_ops": 0,
                "ignored_error_ops": 0,
                "partial_ops": 0,  # operations that had an error, but did do something
                "commands": 0,
            }
            for host in inventory
        }

        # Assign state back references to inventory & config
        inventory.state = config.state = self
        for host in inventory:
            host.state = self

        self.initialised = True

    def to_dict(self):
        return {
            "op_order": self.get_op_order(),
            "ops": self.ops,
            "facts": self.facts,
            "meta": self.meta,
            "results": self.results,
        }

    def add_callback_handler(self, handler):
        if not isinstance(handler, BaseStateCallback):
            raise TypeError(
                ("{0} is not a valid callback handler (use `BaseStateCallback`)").format(handler),
            )
        self.callback_handlers.append(handler)

    def trigger_callbacks(self, method_name: str, *args, **kwargs):
        for handler in self.callback_handlers:
            func = getattr(handler, method_name)
            func(self, *args, **kwargs)

    @contextmanager
    def preserve_loop_order(self, items):
        logger.warning(
            (
                "Using `state.preserve_loop_order` is not longer required for operations to be "
                "executed in correct loop order and can be safely removed."
            ),
        )
        yield lambda: items

    def get_op_order(self):
        ts = TopologicalSorter()

        for host in self.inventory:
            for i, op_hash in enumerate(host.op_hash_order):
                if not i:
                    ts.add(op_hash)
                else:
                    ts.add(op_hash, host.op_hash_order[i - 1])

        final_op_order = []

        try:
            ts.prepare()
        except CycleError as e:
            raise PyinfraError(
                (
                    "Cycle detected in operation ordering DAG.\n"
                    f"    Error: {e}\n\n"
                    "    This can happen when using loops in operation code, "
                    "please see: https://docs.pyinfra.com/en/latest/deploy-process.html#loops-cycle-errors"  # noqa: E501
                ),
            )

        while ts.is_active():
            # Ensure that where we have multiple different operations that can be executed in any
            # dependency order we order them by line numbers.
            node_group = sorted(
                ts.get_ready(),
                key=lambda op_hash: self.op_meta[op_hash]["op_order"],
            )
            ts.done(*node_group)
            final_op_order.extend(node_group)

        return final_op_order

    def get_op_meta(self, op_hash: str):
        return self.op_meta[op_hash]

    def get_op_data(self, host: "Host", op_hash: str):
        return self.ops[host][op_hash]

    def set_op_data(self, host: "Host", op_hash: str, op_data):
        self.ops[host][op_hash] = op_data

    def activate_host(self, host: "Host"):
        """
        Flag a host as active.
        """

        logger.debug("Activating host: %s", host)

        # Add to *both* activated and active - active will reduce as hosts fail
        # but connected will not, enabling us to track failed %.
        self.activated_hosts.add(host)
        self.active_hosts.add(host)

    def fail_hosts(self, hosts_to_fail, activated_count=None):
        """
        Flag a ``set`` of hosts as failed, error for ``config.FAIL_PERCENT``.
        """

        if not hosts_to_fail:
            return

        activated_count = activated_count or len(self.activated_hosts)

        logger.debug(
            "Failing hosts: {0}".format(
                ", ".join(
                    (host.name for host in hosts_to_fail),
                ),
            ),
        )

        self.failed_hosts.update(hosts_to_fail)

        self.active_hosts -= hosts_to_fail

        # Check we're not above the fail percent
        active_hosts = self.active_hosts

        # No hosts left!
        if not active_hosts:
            raise PyinfraError("No hosts remaining!")

        if self.config.FAIL_PERCENT is not None:
            percent_failed = (1 - len(active_hosts) / activated_count) * 100

            if percent_failed > self.config.FAIL_PERCENT:
                raise PyinfraError(
                    "Over {0}% of hosts failed ({1}%)".format(
                        self.config.FAIL_PERCENT,
                        int(round(percent_failed)),
                    ),
                )

    def is_host_in_limit(self, host: "Host"):
        """
        Returns a boolean indicating if the host is within the current state limit.
        """

        limit_hosts = self.limit_hosts

        if not isinstance(limit_hosts, list):
            return True
        return host in limit_hosts

    def get_temp_filename(self, hash_key: Optional[str] = None, hash_filename: bool = True):
        """
        Generate a temporary filename for this deploy.
        """

        if not hash_key:
            hash_key = str(uuid4())

        if hash_filename:
            hash_key = sha1_hash(hash_key)

        return "{0}/pyinfra-{1}".format(self.config.TEMP_DIR, hash_key)



import os
import sys
from datetime import datetime
from io import StringIO
from unittest import TestCase

import pytest

from pyinfra.operations import server
from pyinfra_cli.exceptions import CliError
from pyinfra_cli.util import get_operation_and_args, json_encode


class TestCliUtil(TestCase):
    def test_json_encode_function(self):
        assert json_encode(get_operation_and_args) == "Function: get_operation_and_args"

    def test_json_encode_datetime(self):
        now = datetime.utcnow()
        assert json_encode(now) == now.isoformat()

    def test_json_encode_file(self):
        file = StringIO()
        assert json_encode(file) == "In memory file: "

    def test_json_encode_set(self):
        assert json_encode({1, 2, 3}) == [1, 2, 3]

    def test_setup_no_module(self):
        with self.assertRaises(CliError) as context:
            get_operation_and_args(("no.op",))
        assert context.exception.message == "No such module: no"

    def test_setup_no_op(self):
        with self.assertRaises(CliError) as context:
            get_operation_and_args(("server.no",))

        assert context.exception.message == "No such operation: server.no"

    def test_setup_op_and_args(self):
        commands = ("server.user", "one", "two", "hello=world")

        assert get_operation_and_args(commands) == (
            server.user,
            (["one", "two"], {"hello": "world"}),
        )

    def test_setup_op_and_json_args(self):
        commands = ("server.user", '[["one", "two"], {"hello": "world"}]')

        assert get_operation_and_args(commands) == (
            server.user,
            (["one", "two"], {"hello": "world"}),
        )


@pytest.fixture(scope="function")
def user_sys_path():
    user_pkg = os.path.dirname(__file__) + "/user"
    sys.path.append(user_pkg)
    yield None
    sys.path.pop()
    to_rm = []
    for k, v in sys.modules.items():
        v = getattr(v, "__file__", "")
        if isinstance(v, str) and v.startswith(user_pkg):
            to_rm.append(k)
    for k in to_rm:
        del sys.modules[k]


# def test_no_user_op():
#     commands = ('test_ops.dummy_op', 'arg1', 'arg2')
#     with pytest.raises(CliError, match='^No such module: test_ops$'):
#         get_operation_and_args(commands)


def test_user_op(user_sys_path):
    commands = ("test_ops.dummy_op", "arg1", "arg2")
    res = get_operation_and_args(commands)

    import test_ops

    assert res == (test_ops.dummy_op, (["arg1", "arg2"], {}))



from .command import FileDownloadCommand  # noqa: F401 # pragma: no cover
from .command import (  # noqa: F401
    FileUploadCommand,
    FunctionCommand,
    MaskString,
    QuoteString,
    RsyncCommand,
    StringCommand,
)
from .config import Config  # noqa: F401 # pragma: no cover
from .deploy import deploy  # noqa: F401 # pragma: no cover
from .exceptions import DeployError  # noqa: F401 # pragma: no cover
from .exceptions import (  # noqa: F401
    InventoryError,
    OperationError,
    OperationTypeError,
    OperationValueError,
)
from .facts import FactBase, ShortFactBase  # noqa: F401 # pragma: no cover
from .host import Host  # noqa: F401 # pragma: no cover
from .inventory import Inventory  # noqa: F401 # pragma: no cover
from .operation import operation  # noqa: F401 # pragma: no cover
from .state import BaseStateCallback, State  # noqa: F401 # pragma: no cover



"""
Manage brew packages on mac/OSX. See https://brew.sh/
"""

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.brew import BrewCasks, BrewPackages, BrewTaps, BrewVersion, new_cask_cli

from .util.packaging import ensure_packages


@operation(is_idempotent=False)
def update():
    """
    Updates brew repositories.
    """

    yield "brew update"


_update = update  # noqa: E305


@operation(is_idempotent=False)
def upgrade():
    """
    Upgrades all brew packages.
    """

    yield "brew upgrade"


_upgrade = upgrade  # noqa: E305


@operation
def packages(
    packages=None,
    present=True,
    latest=False,
    update=False,
    upgrade=False,
):
    """
    Add/remove/update brew packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + update: run ``brew update`` before installing packages
    + upgrade: run ``brew upgrade`` before installing packages

    Versions:
        Package versions can be pinned like brew: ``<pkg>@<version>``.

    **Examples:**

    .. code:: python

        # Update package list and install packages
        brew.packages(
            name='Install Vim and vimpager',
            packages=["vimpager", "vim"],
            update=True,
        )

        # Install the latest versions of packages (always check)
        brew.packages(
            name="Install latest Vim",
            packages=["vim"],
            latest=True,
        )
    """

    if update:
        yield from _update()

    if upgrade:
        yield from _upgrade()

    yield from ensure_packages(
        host,
        packages,
        host.get_fact(BrewPackages),
        present,
        install_command="brew install",
        uninstall_command="brew uninstall",
        upgrade_command="brew upgrade",
        version_join="@",
        latest=latest,
    )


def cask_args(host):
    return ("", " --cask") if new_cask_cli(host.get_fact(BrewVersion)) else ("cask ", "")


@operation(
    is_idempotent=False,
    pipeline_facts={"brew_version": ""},
)
def cask_upgrade():
    """
    Upgrades all brew casks.
    """

    yield "brew %supgrade%s" % cask_args(host)


@operation(
    pipeline_facts={"brew_version": ""},
)
def casks(
    casks=None,
    present=True,
    latest=False,
    upgrade=False,
):
    """
    Add/remove/update brew casks.

    + casks: list of casks to ensure
    + present: whether the casks should be installed
    + latest: whether to upgrade casks without a specified version
    + upgrade: run brew cask upgrade before installing casks

    Versions:
        Cask versions can be pinned like brew: ``<pkg>@<version>``.

    **Example:**

    .. code:: python

        brew.casks(
            name='Upgrade and install the latest cask',
            casks=["godot"],
            upgrade=True,
            latest=True,
        )

    """

    if upgrade:
        yield from cask_upgrade()

    args = cask_args(host)

    yield from ensure_packages(
        host,
        casks,
        host.get_fact(BrewCasks),
        present,
        install_command="brew %sinstall%s" % args,
        uninstall_command="brew %suninstall%s" % args,
        upgrade_command="brew %supgrade%s" % args,
        version_join="@",
        latest=latest,
    )


@operation
def tap(src, present=True):
    """
    Add/remove brew taps.

    + src: the name of the tap
    + present: whether this tap should be present or not

    **Examples:**

    .. code:: python

        brew.tap(
            name="Add a brew tap",
            src="includeos/includeos",
        )

        # Multiple taps
        for tap in ["includeos/includeos", "ktr0731/evans"]:
            brew.tap(
                name={f"Add brew tap {tap}"},
                src=tap,
            )

    """

    taps = host.get_fact(BrewTaps)
    is_tapped = src in taps

    if present:
        if is_tapped:
            host.noop("tap {0} already exists".format(src))
        else:
            yield "brew tap {0}".format(src)
            taps.append(src)

    elif not present:
        if is_tapped:
            yield "brew untap {0}".format(src)
            taps.remove(src)
        else:
            host.noop("tap {0} does not exist".format(src))



from pyinfra.api import FactBase

from .util.packaging import parse_packages

CHOCO_REGEX = r"^([a-zA-Z0-9\.\-\+\_]+)\s([0-9\.]+)$"


class ChocoPackages(FactBase):
    """
    Returns a dict of installed choco (Chocolatey) packages:

    .. code:: python

        {
            "package_name": ["version"],
        }
    """

    command = "choco list --local-only"
    shell_executable = "ps"

    default = dict

    def process(self, output):
        return parse_packages(CHOCO_REGEX, output)


class ChocoVersion(FactBase):
    """
    Returns the choco (Chocolatey) version.
    """

    command = "choco --version"

    @staticmethod
    def process(output):
        return "".join(output).replace("\n", "")



"""
Manage pacman packages. (Arch Linux package manager)
"""

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.pacman import PacmanPackages, PacmanUnpackGroup

from .util.packaging import ensure_packages


@operation(is_idempotent=False)
def upgrade():
    """
    Upgrades all pacman packages.
    """

    yield "pacman --noconfirm -Su"


_upgrade = upgrade  # noqa: E305


@operation(is_idempotent=False)
def update():
    """
    Updates pacman repositories.
    """

    yield "pacman -Sy"


_update = update  # noqa: E305


@operation
def packages(
    packages=None,
    present=True,
    update=False,
    upgrade=False,
):
    """
    Add/remove pacman packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + update: run ``pacman -Sy`` before installing packages
    + upgrade: run ``pacman -Su`` before installing packages

    Versions:
        Package versions can be pinned like pacman: ``<pkg>=<version>``.

    **Example:**

    .. code:: python

        pacman.packages(
            name="Install Vim and a plugin",
            packages=["vim-fugitive", "vim"],
            update=True,
        )
    """

    if update:
        yield from _update()

    if upgrade:
        yield from _upgrade()

    yield from ensure_packages(
        host,
        packages,
        host.get_fact(PacmanPackages),
        present,
        install_command="pacman --noconfirm -S",
        uninstall_command="pacman --noconfirm -R",
        expand_package_fact=lambda package: host.get_fact(
            PacmanUnpackGroup,
            name=package,
        ),
    )



"""
Manage systemd services.
"""

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.systemd import SystemdEnabled, SystemdStatus, _make_systemctl_cmd

from .util.service import handle_service_control


@operation(is_idempotent=False)
def daemon_reload(user_mode=False, machine=None, user_name=None):
    """
    Reload the systemd daemon to read unit file changes.

    + user_mode: whether to use per-user systemd (systemctl --user) or not
    + machine: the machine name to connect to
    + user_name: connect to a specific user's systemd session
    """

    systemctl_cmd = _make_systemctl_cmd(
        user_mode=user_mode,
        machine=machine,
        user_name=user_name,
    )

    yield "{0} daemon-reload".format(systemctl_cmd)


_daemon_reload = daemon_reload  # noqa: E305


@operation
def service(
    service,
    running=True,
    restarted=False,
    reloaded=False,
    command=None,
    enabled=None,
    daemon_reload=False,
    user_mode=False,
    machine=None,
    user_name=None,
):
    """
    Manage the state of systemd managed units.

    + service: name of the systemd unit to manage
    + running: whether the unit should be running
    + restarted: whether the unit should be restarted
    + reloaded: whether the unit should be reloaded
    + command: custom command to pass like: ``/etc/rc.d/<name> <command>``
    + enabled: whether this unit should be enabled/disabled on boot
    + daemon_reload: reload the systemd daemon to read updated unit files
    + user_mode: whether to use per-user systemd (systemctl --user) or not
    + machine: the machine name to connect to
    + user_name: connect to a specific user's systemd session

    **Examples:**

    .. code:: python

        systemd.service(
            name="Restart and enable the dnsmasq service",
            service="dnsmasq.service",
            running=True,
            restarted=True,
            enabled=True,
        )

        systemd.service(
            name="Enable logrotate timer",
            service="logrotate.timer",
            running=True,
            enabled=True,
        )

    """

    systemctl_cmd = _make_systemctl_cmd(
        user_mode=user_mode,
        machine=machine,
        user_name=user_name,
    )

    if "." not in service:
        service = "{0}.service".format(service)

    if daemon_reload:
        yield from _daemon_reload(
            user_mode=user_mode,
            machine=machine,
            user_name=user_name,
        )

    yield from handle_service_control(
        host,
        service,
        host.get_fact(
            SystemdStatus,
            user_mode=user_mode,
            machine=machine,
            user_name=user_name,
        ),
        " ".join([systemctl_cmd, "{1}", "{0}"]),
        running,
        restarted,
        reloaded,
        command,
    )

    if isinstance(enabled, bool):
        systemd_enabled = host.get_fact(
            SystemdEnabled,
            user_mode=user_mode,
            machine=machine,
            user_name=user_name,
        )
        is_enabled = systemd_enabled.get(service, False)

        # Isn't enabled and want enabled?
        if not is_enabled and enabled is True:
            yield "{0} enable {1}".format(systemctl_cmd, service)
            systemd_enabled[service] = True

        # Is enabled and want disabled?
        elif is_enabled and enabled is False:
            yield "{0} disable {1}".format(systemctl_cmd, service)
            systemd_enabled[service] = False



"""
The pyinfra facts API. Facts enable pyinfra to collect remote server state which
is used to "diff" with the desired state, producing the final commands required
for a deploy.
"""

import re
from inspect import getcallargs
from socket import error as socket_error, timeout as timeout_error
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Type, Union

import click
import gevent
from paramiko import SSHException

from pyinfra import logger
from pyinfra.api import StringCommand
from pyinfra.api.arguments import pop_global_arguments
from pyinfra.api.util import (
    get_kwargs_str,
    log_error_or_warning,
    log_host_command_error,
    make_hash,
    print_host_combined_output,
)
from pyinfra.connectors.util import split_combined_output
from pyinfra.context import ctx_host, ctx_state
from pyinfra.progress import progress_spinner

from .arguments import get_executor_kwarg_keys

if TYPE_CHECKING:
    from pyinfra.api.host import Host
    from pyinfra.api.state import State

SUDO_REGEX = r"^sudo: unknown user:"
SU_REGEXES = (
    r"^su: user .+ does not exist",
    r"^su: unknown login",
)


class FactNameMeta(type):
    def __init__(cls, name: str, bases, attrs):
        module_name = cls.__module__.replace("pyinfra.facts.", "")
        cls.name = f"{module_name}.{cls.__name__}"


class FactBase(object, metaclass=FactNameMeta):
    name: str

    abstract: bool = True

    shell_executable: Optional[str] = None

    requires_command: Optional[str] = None

    command: Union[str, Callable]

    @staticmethod
    def default():
        """
        Set the default attribute to be a type (eg list/dict).
        """

    @staticmethod
    def process(output):
        return "\n".join(output)

    def process_pipeline(self, args, output):
        return {arg: self.process([output[i]]) for i, arg in enumerate(args)}


class ShortFactBase(object, metaclass=FactNameMeta):
    fact: Type[FactBase]


def get_short_facts(state: "State", host: "Host", short_fact, **kwargs):
    fact_data = get_fact(state, host, short_fact.fact, **kwargs)
    return short_fact().process_data(fact_data)


def _make_command(command_attribute, host_args):
    if callable(command_attribute):
        host_args.pop("self", None)
        return command_attribute(**host_args)
    return command_attribute


def _get_executor_kwargs(
    state: "State",
    host: "Host",
    override_kwargs: Optional[Dict[str, Any]] = None,
    override_kwarg_keys: Optional[List[str]] = None,
):
    if override_kwargs is None:
        override_kwargs = {}
    if override_kwarg_keys is None:
        override_kwarg_keys = []

    # Apply any current op kwargs that *weren't* found in the overrides
    override_kwargs.update(
        {
            key: value
            for key, value in (host.current_op_global_kwargs or {}).items()
            if key not in override_kwarg_keys
        },
    )

    return {
        key: value for key, value in override_kwargs.items() if key in get_executor_kwarg_keys()
    }


def get_facts(state: "State", *args, **kwargs):
    def get_fact_with_context(state, host, *args, **kwargs):
        with ctx_state.use(state):
            with ctx_host.use(host):
                return get_fact(state, host, *args, **kwargs)

    greenlet_to_host = {
        state.pool.spawn(get_fact_with_context, state, host, *args, **kwargs): host
        for host in state.inventory.iter_active_hosts()
    }

    results = {}

    with progress_spinner(greenlet_to_host.values()) as progress:
        for greenlet in gevent.iwait(greenlet_to_host.keys()):
            host = greenlet_to_host[greenlet]
            results[host] = greenlet.get()
            progress(host)

    return results


def get_fact(
    state: "State",
    host: "Host",
    cls: Type[FactBase],
    args: Optional[Any] = None,
    kwargs: Optional[Any] = None,
    ensure_hosts: Optional[Any] = None,
    apply_failed_hosts: bool = True,
    fact_hash: Optional[Any] = None,
    use_cache: bool = True,
):
    if issubclass(cls, ShortFactBase):
        return get_short_facts(
            state,
            host,
            cls,
            args=args,
            kwargs=kwargs,
            ensure_hosts=ensure_hosts,
        )

    with host.facts_lock:
        if use_cache and fact_hash and fact_hash in host.facts:
            return host.facts[fact_hash]

        return _get_fact(
            state,
            host,
            cls,
            args,
            kwargs,
            ensure_hosts,
            apply_failed_hosts,
            fact_hash,
        )


def _get_fact(
    state: "State",
    host: "Host",
    cls: Type[FactBase],
    args: Optional[List] = None,
    kwargs: Optional[Dict] = None,
    ensure_hosts: Optional[Any] = None,
    apply_failed_hosts: bool = True,
    fact_hash: Optional[Any] = None,
):
    fact = cls()
    name = fact.name

    args = args or []
    kwargs = kwargs or {}

    # Get the defaults *and* overrides by popping from kwargs, executor kwargs passed
    # into get_fact override everything else (applied below).
    override_kwargs, override_kwarg_keys = pop_global_arguments(
        kwargs,
        state=state,
        host=host,
        keys_to_check=get_executor_kwarg_keys(),
    )

    executor_kwargs = _get_executor_kwargs(
        state,
        host,
        override_kwargs=override_kwargs,
        override_kwarg_keys=override_kwarg_keys,
    )

    if args or kwargs:
        assert not isinstance(fact.command, str)
        # Merges args & kwargs into a single kwargs dictionary
        kwargs = getcallargs(fact.command, *args, **kwargs)

    kwargs_str = get_kwargs_str(kwargs)
    logger.debug(
        "Getting fact: %s (%s) (ensure_hosts: %r)",
        name,
        kwargs_str,
        ensure_hosts,
    )

    if not host.connected:
        host.connect(
            reason=f"to load fact: {name} ({kwargs_str})",
            raise_exceptions=True,
        )

    ignore_errors = (host.current_op_global_kwargs or {}).get(
        "ignore_errors",
        state.config.IGNORE_ERRORS,
    )

    # Facts can override the shell (winrm powershell vs cmd support)
    if fact.shell_executable:
        executor_kwargs["shell_executable"] = fact.shell_executable

    command = _make_command(fact.command, kwargs)
    requires_command = _make_command(fact.requires_command, kwargs)
    if requires_command:
        command = StringCommand(
            # Command doesn't exist, return 0 *or* run & return fact command
            "!",
            "command",
            "-v",
            requires_command,
            ">/dev/null",
            "||",
            command,
        )

    status = False
    stdout = []
    combined_output_lines = []

    try:
        status, combined_output_lines = host.run_shell_command(
            command,
            print_output=state.print_fact_output,
            print_input=state.print_fact_input,
            return_combined_output=True,
            **executor_kwargs,
        )
    except (timeout_error, socket_error, SSHException) as e:
        log_host_command_error(
            host,
            e,
            timeout=executor_kwargs["timeout"],
        )

    stdout, stderr = split_combined_output(combined_output_lines)

    data = fact.default()

    if status:
        if stdout:
            data = fact.process(stdout)
    elif stderr:
        # If we have error output and that error is sudo or su stating the user
        # does not exist, do not fail but instead return the default fact value.
        # This allows for users that don't currently but may be created during
        # other operations.
        first_line = stderr[0]
        if executor_kwargs["sudo_user"] and re.match(SUDO_REGEX, first_line):
            status = True
        if executor_kwargs["su_user"] and any(re.match(regex, first_line) for regex in SU_REGEXES):
            status = True

    if status:
        log_message = "{0}{1}".format(
            host.print_prefix,
            "Loaded fact {0}{1}".format(
                click.style(name, bold=True),
                f" ({get_kwargs_str(kwargs)})" if kwargs else "",
            ),
        )
        if state.print_fact_info:
            logger.info(log_message)
        else:
            logger.debug(log_message)
    else:
        if not state.print_fact_output:
            print_host_combined_output(host, combined_output_lines)

        log_error_or_warning(
            host,
            ignore_errors,
            description=("could not load fact: {0} {1}").format(name, get_kwargs_str(kwargs)),
        )

    # Check we've not failed
    if not status and not ignore_errors and apply_failed_hosts:
        state.fail_hosts({host})

    if fact_hash:
        host.facts[fact_hash] = data
    return data


def _get_fact_hash(state: "State", host: "Host", cls, args, kwargs):
    args = args or None
    kwargs = kwargs or None
    return make_hash((cls, args, kwargs, _get_executor_kwargs(state, host)))


def get_host_fact(
    state: "State",
    host: "Host",
    cls,
    args: Optional[List] = None,
    kwargs: Optional[Dict] = None,
):
    fact_hash = _get_fact_hash(state, host, cls, args, kwargs)
    return get_fact(state, host, cls, args=args, kwargs=kwargs, fact_hash=fact_hash)


def reload_host_fact(
    state: "State",
    host: "Host",
    cls,
    args: Optional[List] = None,
    kwargs: Optional[Dict] = None,
):
    fact_hash = _get_fact_hash(state, host, cls, args, kwargs)
    return get_fact(
        state,
        host,
        cls,
        args=args,
        kwargs=kwargs,
        fact_hash=fact_hash,
        use_cache=False,
    )


def create_host_fact(
    state: "State",
    host: "Host",
    cls,
    data,
    args: Optional[List] = None,
    kwargs: Optional[Dict] = None,
):
    fact_hash = _get_fact_hash(state, host, cls, args, kwargs)
    host.facts[fact_hash] = data


def delete_host_fact(
    state: "State",
    host: "Host",
    cls,
    args: Optional[List] = None,
    kwargs: Optional[Dict] = None,
):
    fact_hash = _get_fact_hash(state, host, cls, args, kwargs)
    host.facts.pop(fact_hash, None)



from unittest import TestCase

from pyinfra.api.util import format_exception, get_caller_frameinfo, try_int


class TestApiUtil(TestCase):
    def test_try_int_number(self):
        assert try_int("1") == 1

    def test_try_int_fail(self):
        assert try_int("hello") == "hello"

    def test_get_caller_frameinfo(self):
        def _get_caller_frameinfo():
            return get_caller_frameinfo()

        frameinfo = _get_caller_frameinfo()
        assert frameinfo.lineno == 17  # called by the line above

    def test_format_exception(self):
        exception = Exception("I am a message", 1)
        assert format_exception(exception) == "Exception('I am a message', 1)"



# This file only exists to support:
# from pyinfra import operations
# operations.X.Y

from glob import glob
from os import path

module_filenames = glob(path.join(path.dirname(__file__), "*.py"))
module_names = [path.basename(name)[:-3] for name in module_filenames]
__all__ = [name for name in module_names if name != "__init__"]

from . import *  # noqa



from unittest import TestCase
from unittest.mock import mock_open, patch

from paramiko import ProxyCommand

from pyinfra.connectors.sshuserclient import SSHClient
from pyinfra.connectors.sshuserclient.client import AskPolicy, get_ssh_config

SSH_CONFIG_DATA = """
# Comment
Host 127.0.0.1
    IdentityFile /id_rsa
    IdentityFile /id_rsa2
    User testuser
    Port 33
    ProxyCommand ssh thing

Include other_file
"""

SSH_CONFIG_OTHER_FILE = """
Host 192.168.1.1
    User "otheruser"
    ProxyCommand None
    ForwardAgent yes
    UserKnownHostsFile ~/.ssh/test3
"""

SSH_CONFIG_OTHER_FILE_PROXYJUMP = """
Host 192.168.1.2
    User "otheruser"
    ProxyJump nottestuser@127.0.0.1
    ForwardAgent yes
"""

BAD_SSH_CONFIG_DATA = """
&
"""

LOOPING_SSH_CONFIG_DATA = """
Include other_file
"""


class TestSSHUserConfigMissing(TestCase):
    def setUp(self):
        get_ssh_config.cache = {}

    @patch(
        "pyinfra.connectors.sshuserclient.client.path.exists",
        lambda path: False,
    )
    def test_load_ssh_config_no_exist(self):
        client = SSHClient()

        _, config, forward_agent, missing_host_key_policy, host_keys_file = client.parse_config(
            "127.0.0.1",
        )

        assert config.get("port") == 22


@patch(
    "pyinfra.connectors.sshuserclient.client.path.exists",
    lambda path: True,
)
@patch(
    "pyinfra.connectors.sshuserclient.config.glob.iglob",
    lambda path: ["other_file"],
)
@patch(
    "pyinfra.connectors.sshuserclient.config.path.isfile",
    lambda path: True,
)
@patch(
    "pyinfra.connectors.sshuserclient.config.path.expanduser",
    lambda path: path,
)
@patch(
    "pyinfra.connectors.sshuserclient.config.path.isabs",
    lambda path: True,
)
@patch(
    "paramiko.config.LazyFqdn.__str__",
    lambda self: "",
)
class TestSSHUserConfig(TestCase):
    def setUp(self):
        get_ssh_config.cache = {}

    @patch(
        "pyinfra.connectors.sshuserclient.client.open",
        mock_open(read_data=SSH_CONFIG_DATA),
        create=True,
    )
    @patch(
        "pyinfra.connectors.sshuserclient.config.open",
        mock_open(read_data=SSH_CONFIG_OTHER_FILE),
        create=True,
    )
    def test_load_ssh_config(self):
        client = SSHClient()

        _, config, forward_agent, missing_host_key_policy, host_keys_file = client.parse_config(
            "127.0.0.1",
        )

        assert config.get("key_filename") == ["/id_rsa", "/id_rsa2"]
        assert config.get("username") == "testuser"
        assert config.get("port") == 33
        assert isinstance(config.get("sock"), ProxyCommand)
        assert forward_agent is False
        assert isinstance(missing_host_key_policy, AskPolicy)
        assert host_keys_file == "~/.ssh/known_hosts"  # OpenSSH default

        (
            _,
            other_config,
            forward_agent,
            missing_host_key_policy,
            host_keys_file,
        ) = client.parse_config("192.168.1.1")

        assert other_config.get("username") == "otheruser"
        assert forward_agent is True
        assert isinstance(missing_host_key_policy, AskPolicy)
        assert host_keys_file == "~/.ssh/test3"

    @patch(
        "pyinfra.connectors.sshuserclient.client.open",
        mock_open(read_data=BAD_SSH_CONFIG_DATA),
        create=True,
    )
    def test_invalid_ssh_config(self):
        client = SSHClient()

        with self.assertRaises(Exception) as context:
            client.parse_config("127.0.0.1")

        assert context.exception.args[0] == "Unparsable line &"

    @patch(
        "pyinfra.connectors.sshuserclient.client.open",
        mock_open(read_data=LOOPING_SSH_CONFIG_DATA),
        create=True,
    )
    @patch(
        "pyinfra.connectors.sshuserclient.config.open",
        mock_open(read_data=LOOPING_SSH_CONFIG_DATA),
        create=True,
    )
    def test_include_loop_ssh_config(self):
        client = SSHClient()

        with self.assertRaises(Exception) as context:
            client.parse_config("127.0.0.1")

        assert context.exception.args[0] == "Include loop detected in ssh config file: other_file"

    @patch(
        "pyinfra.connectors.sshuserclient.client.open",
        mock_open(read_data=SSH_CONFIG_DATA),
        create=True,
    )
    @patch(
        "pyinfra.connectors.sshuserclient.config.open",
        mock_open(read_data=SSH_CONFIG_OTHER_FILE_PROXYJUMP),
        create=True,
    )
    @patch("pyinfra.connectors.sshuserclient.SSHClient.connect")
    @patch("pyinfra.connectors.sshuserclient.SSHClient.gateway")
    def test_load_ssh_config_proxyjump(self, fake_gateway, fake_ssh_connect):
        client = SSHClient()

        # Load the SSH config with ProxyJump configured
        _, config, forward_agent, _, _ = client.parse_config(
            "192.168.1.2",
            {"port": 1022},
            ssh_config_file="other_file",
        )

        fake_ssh_connect.assert_called_once_with(
            "127.0.0.1",
            _pyinfra_ssh_config_file="other_file",
            port="33",
            sock=None,
            username="nottestuser",
        )
        fake_gateway.assert_called_once_with("192.168.1.2", 1022, "192.168.1.2", 1022)

    @patch("pyinfra.connectors.sshuserclient.client.open", mock_open(), create=True)
    @patch("pyinfra.connectors.sshuserclient.client.ParamikoClient.connect")
    def test_test_paramiko_connect_kwargs(self, fake_paramiko_connect):
        client = SSHClient()
        client.connect("hostname", _pyinfra_ssh_paramiko_connect_kwargs={"test": "kwarg"})

        fake_paramiko_connect.assert_called_once_with(
            "hostname",
            port=22,
            test="kwarg",
        )



from unittest import TestCase

from pyinfra.api.host import HostData


class TestHostData(TestCase):
    def test_host_data(self):
        data = HostData("somehost", {"hello": "world"})
        assert data.hello == "world"
        assert data.get("hello") == "world"

    def test_host_data_multiple(self):
        data = HostData(
            "somehost",
            {"hello": "world"},
            {"hello": "not-world", "another": "thing"},
        )
        assert data.hello == "world"
        assert data.get("hello") == "world"
        assert data.another == "thing"

    def test_host_data_override(self):
        data = HostData("somehost", {"hello": "world"})
        assert data.hello == "world"

        data.hello = "override-world"
        assert data.hello == "override-world"

    def test_host_data_missing(self):
        data = HostData("somehost", {"hello": "world"})

        with self.assertRaises(AttributeError) as context:
            getattr(data, "not-a-key")

        assert context.exception.args[0] == "Host `somehost` has no data `not-a-key`"
        assert data.get("not-a-key") is None



from pyinfra.api.facts import FactBase

from .util.win_files import parse_win_ls_output


class File(FactBase):
    # Types must match WIN_FLAG_TO_TYPE in .util.win_files.py
    type = "file"
    shell_executable = "ps"

    def command(self, path):
        self.path = path
        return ('if (Test-Path "{0}") {{ Get-ItemProperty -Path "{0}" }}').format(path)

    def process(self, output):
        if len(output) < 7:
            return None

        # Note: The first 7 lines are header lines
        return parse_win_ls_output(output[7], self.type)


class Link(File):
    # Types must match WIN_FLAG_TO_TYPE in .util.win_files.py
    type = "link"


class Directory(File):
    # Types must match WIN_FLAG_TO_TYPE in .util.win_files.py
    type = "directory"


class TempDir(FactBase):
    # Types must match WIN_FLAG_TO_TYPE in .util.win_files.py
    type = "directory"
    shell_executable = "ps"

    def command(self):
        return "[System.IO.Path]::GetTempPath()"

    def process(self, output):
        return output[0]


class Sha1File(FactBase):
    """
    Returns a SHA1 hash of a file.
    """

    shell_executable = "ps"

    def command(self, path):
        return (
            'if (Test-Path "{0}") {{ ' '(Get-FileHash -Algorithm SHA1 "{0}").hash' " }}"
        ).format(path)

    def process(self, output):
        return output[0] if not output else None


class Sha256File(FactBase):
    """
    Returns a SHA256 hash of a file.
    """

    shell_executable = "ps"

    def command(self, path):
        return (
            'if (Test-Path "{0}") {{ ' '(Get-FileHash -Algorithm SHA256 "{0}").hash' " }}"
        ).format(path)

    def process(self, output):
        return output[0] if len(output[0]) > 0 else None


class Md5File(FactBase):
    """
    Returns an MD5 hash of a file.
    """

    shell_executable = "ps"

    def command(self, path):
        return ('if (Test-Path "{0}") {{ ' '(Get-FileHash -Algorithm MD5 "{0}").hash' " }}").format(
            path,
        )

    def process(self, output):
        return output[0] if len(output[0]) > 0 else None



import re
import sys
from io import open

try:
    from setuptools import find_packages, setup

except ImportError:
    print(
        """
Error: pyinfra needs setuptools in order to install:

using pip: pip install setuptools
using a package manager (apt, yum, etc), normally named: python-setuptools
    """.strip(),
    )

    sys.exit(1)


INSTALL_REQUIRES = (
    "gevent>=1.5",
    "paramiko>=2.7,<3",  # 2.7 (2019) adds OpenSSH key format + Match SSH config
    "click>2",
    "colorama<1",  # Windows color support for click
    "jinja2>2,<4",
    "python-dateutil>2,<3",
    "setuptools",
    "configparser",
    "pywinrm",
    "distro>=1.5,<2",
    # Backport of graphlib used for DAG operation ordering
    'graphlib_backport ; python_version < "3.9"',
)

ANSIBLE_REQUIRES = ("pyyaml",)  # extras for parsing Ansible inventory

TEST_REQUIRES = ANSIBLE_REQUIRES + (
    # Unit testing
    "pytest==7.0.1",
    "pytest-cov==3.0.0",
    "coverage==6.2",
    "codecov==2.1.12",
    # Formatting & linting
    "black==22.3.0",
    "isort==5.10.1",
    "flake8==4.0.1",
    "flake8-black==0.3.0",
    "flake8-isort==4.1.1",
    # Typing
    "mypy==0.971",
    "types-cryptography",
    "types-paramiko",
    "types-python-dateutil",
    "types-PyYAML",
    "types-setuptools",
)

DOCS_REQUIRES = (
    "pyinfra-guzzle_sphinx_theme==0.14",
    "recommonmark==0.5.0",
    "sphinx==2.2.1",
    # Pinned to fix: https://github.com/sphinx-doc/sphinx/issues/9727
    "docutils==0.17.1",
)

DEV_REQUIRES = (
    TEST_REQUIRES
    + DOCS_REQUIRES
    + (
        # Releasing
        "wheel",
        "twine",
        # Dev debugging
        "ipython",
        "ipdb",
        "ipdbplugin",
        # Lint spellchecking, dev only (don't fail CI)
        "flake8-spellcheck==0.12.1",
        "redbaron",  # for generating type stubs
    )
)


def get_version_from_changelog():
    # Regex matching pattern followed by 3 numerical values separated by '.'
    pattern = re.compile(r"^# v(?P<version>[0-9]+\.[0-9]+(\.[0-9]+)?(\.?[a-z0-9]+)?)$")

    with open("CHANGELOG.md", "r") as fn:
        for line in fn.readlines():
            match = pattern.match(line.strip())
            if match:
                return "".join(match.group("version"))
    raise RuntimeError("No version found in CHANGELOG.md")


def get_readme_contents():
    with open("README.md", "r", encoding="utf-8") as f:
        return f.read()


if __name__ == "__main__":
    setup(
        version=get_version_from_changelog(),
        name="pyinfra",
        description="pyinfra automates/provisions/manages/deploys infrastructure.",
        long_description=get_readme_contents(),
        long_description_content_type="text/markdown",
        author="Nick / Fizzadar",
        author_email="pointlessrambler@gmail.com",
        license="MIT",
        url="https://pyinfra.com",
        project_urls={
            "Documentation": "https://docs.pyinfra.com",
            "GitHub": "https://github.com/Fizzadar/pyinfra",
        },
        packages=find_packages(exclude=["tests", "docs"]),
        entry_points={
            "console_scripts": ("pyinfra=pyinfra_cli.__main__:execute_pyinfra",),
            "pyinfra.connectors": [
                "ansible = pyinfra.connectors.ansible",
                "chroot = pyinfra.connectors.chroot",
                "docker = pyinfra.connectors.docker",
                "local = pyinfra.connectors.local",
                "mech = pyinfra.connectors.mech",
                "ssh = pyinfra.connectors.ssh",
                "dockerssh = pyinfra.connectors.dockerssh",
                "vagrant = pyinfra.connectors.vagrant",
                "winrm = pyinfra.connectors.winrm",
                "terraform = pyinfra.connectors.terraform",
            ],
        },
        install_requires=INSTALL_REQUIRES,
        extras_require={
            "test": TEST_REQUIRES,
            "docs": DOCS_REQUIRES,
            "dev": DEV_REQUIRES,
            "ansible": ANSIBLE_REQUIRES,
        },
        include_package_data=True,
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Environment :: Console",
            "Intended Audience :: Developers",
            "Intended Audience :: System Administrators",
            "Intended Audience :: Information Technology",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Topic :: System :: Systems Administration",
            "Topic :: System :: Installation/Setup",
            "Topic :: Utilities",
        ],
    )



"""
**Note**: this connector is a work in progress! While it parses the list of
hosts OK, it doesn't handle nested groups properly yet.

The `@ansible` connector can be used to parse Ansible inventory files.

.. code:: python

    # Load an Ansible inventory relative to the current directory
    pyinfra @ansible/path/to/inventory

    # Load using an absolute path
    pyinfra @ansible//absolute/path/to/inventory
"""
import json
import re
from collections import defaultdict
from configparser import ConfigParser
from os import path
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from pyinfra.api.host import Host

try:
    import yaml
except ImportError:
    yaml = None  # type: ignore

from pyinfra import logger
from pyinfra.api.exceptions import InventoryError
from pyinfra.api.util import memoize


@memoize
def show_warning():
    logger.warning("The @ansible connector is in alpha!")


def make_names_data(inventory_filename: Optional[str] = None):
    show_warning()

    if not inventory_filename:
        raise InventoryError("No Ansible inventory filename provided!")

    if not path.exists(inventory_filename):
        raise InventoryError(
            ("Could not find Ansible inventory file: {0}").format(inventory_filename),
        )

    return parse_inventory(inventory_filename)


def parse_inventory(inventory_filename: str):
    # fallback to INI if no extension
    extension = inventory_filename.split(".")[-1] if "." in inventory_filename else "ini"

    # host:set(groups) mapping
    host_to_groups = {}

    if extension in ["ini"]:
        host_to_groups = parse_inventory_ini(inventory_filename)
    elif extension in ["json"]:
        with open(inventory_filename) as inventory_file:
            inventory_tree = json.load(inventory_file)
            # close file early
        host_to_groups = parse_inventory_tree(inventory_tree)
    elif extension in ["yaml", "yml"]:
        if yaml is None:
            raise Exception(
                (
                    "To parse YAML Ansible inventories requires `pyyaml`. "
                    "Install it with `pip install pyyaml`."
                ),
            )
        with open(inventory_filename) as inventory_file:
            inventory_tree = yaml.safe_load(inventory_file)
            # close file early
        host_to_groups = parse_inventory_tree(inventory_tree)
    else:
        raise InventoryError(("Ansible inventory file format not supported: {0}").format(extension))

    return [(host, {}, sorted(list(host_to_groups.get(host, [])))) for host in host_to_groups]


def parse_inventory_ini(inventory_filename: str):
    config = ConfigParser(
        delimiters=(" "),  # we only handle the hostnames for now
        allow_no_value=True,  # we don't by default have = values
        interpolation=None,  # remove any Python interpolation
    )
    config.read(inventory_filename)

    host_to_groups = defaultdict(set)
    group_to_hosts = defaultdict(set)
    hosts = []

    # First pass - load hosts/groups of hosts
    for section in config.sections():
        if ":" in section:  # ignore :children and :vars sections this time
            continue

        options = config.options(section)
        for host in _parse_ansible_hosts(options):
            hosts.append(host)
            host_to_groups[host].add(section)
            group_to_hosts[section].add(host)

    # Second pass - load any children groups
    for section in config.sections():
        if not section.endswith(":children"):  # we only support :children for now
            continue

        group_name = section.replace(":children", "")

        options = config.options(section)
        for sub_group_name in options:
            sub_group_hosts = group_to_hosts[sub_group_name]
            for host in sub_group_hosts:
                host_to_groups[host].add(group_name)

    return host_to_groups


def _parse_ansible_hosts(hosts):
    for host in hosts:
        expand_match = re.search(r"\[[0-9:]+\]", host)
        if expand_match:
            expand_string = host[expand_match.start() : expand_match.end()]
            bits = expand_string[1:-1].split(":")  # remove the [] either side

            zfill = 0
            if bits[0].startswith("0"):
                zfill = len(bits[0])

            start, end = int(bits[0]), int(bits[1])
            step = int(bits[2]) if len(bits) > 2 else 1

            for n in range(start, end + 1, step):
                number_as_string = "{0}".format(n)
                if zfill:
                    number_as_string = number_as_string.zfill(zfill)

                hostname = host.replace(expand_string, number_as_string)
                yield hostname
        else:
            yield host


def parse_inventory_tree(inventory_tree, host_to_groups=dict(), group_stack=set()):
    for group in inventory_tree:
        # set logic adds tolerance for duplicate group names
        groups = group_stack.union({group})

        if "hosts" in inventory_tree[group]:
            for host in inventory_tree[group]["hosts"]:
                append_groups_to_host(host, groups, host_to_groups)

        if "children" in inventory_tree[group]:
            # recursively parse inventory tree
            parse_inventory_tree(inventory_tree[group]["children"], host_to_groups, groups)

    return host_to_groups


def append_groups_to_host(host: "Host", groups, host_to_groups):
    if host in host_to_groups:
        # set logic handles de-duplication
        host_to_groups[host] = host_to_groups[host].union(groups)
    else:
        host_to_groups[host] = groups



import re

from pyinfra.api import FactBase, ShortFactBase


class Cpus(FactBase):
    """
    Returns the number of CPUs on this server.
    """

    command = "getconf NPROCESSORS_ONLN 2> /dev/null || getconf _NPROCESSORS_ONLN"

    @staticmethod
    def process(output):
        try:
            return int(output[0])
        except ValueError:
            pass


class Memory(FactBase):
    """
    Returns the memory installed in this server, in MB.
    """

    command = "vmstat -s"
    requires_command = "vmstat"

    @staticmethod
    def process(output):
        data = {}

        for line in output:
            line = line.strip()
            value, key = line.split(" ", 1)

            try:
                value = int(value)
            except ValueError:
                continue

            data[key.strip()] = value

        # Easy - Linux just gives us the number
        total_memory = data.get("K total memory", data.get("total memory"))

        # BSD - calculate the total from the # pages and the page size
        if not total_memory:
            bytes_per_page = data.get("bytes per page")
            pages_managed = data.get("pages managed")

            if bytes_per_page and pages_managed:
                total_memory = (pages_managed * bytes_per_page) / 1024

        if total_memory:
            return int(round(total_memory / 1024))


class BlockDevices(FactBase):
    """
    Returns a dict of (mounted) block devices:

    .. code:: python

        {
            "/dev/sda1": {
                "available": "39489508",
                "used_percent": "3",
                "mount": "/",
                "used": "836392",
                "blocks": "40325900"
            },
        }
    """

    command = "df"
    regex = r"([a-zA-Z0-9\/\-_]+)\s+([0-9]+)\s+([0-9]+)\s+([0-9]+)\s+([0-9]{1,3})%\s+([a-zA-Z\/0-9\-_]+)"  # noqa: E501
    default = dict

    def process(self, output):
        devices = {}

        for line in output:
            matches = re.match(self.regex, line)
            if matches:
                if matches.group(1) == "none":
                    continue

                devices[matches.group(1)] = {
                    "blocks": matches.group(2),
                    "used": matches.group(3),
                    "available": matches.group(4),
                    "used_percent": matches.group(5),
                    "mount": matches.group(6),
                }

        return devices


nettools_1_regexes = [
    (
        r"^inet addr:([0-9\.]+).+Bcast:([0-9\.]+).+Mask:([0-9\.]+)$",
        ("ipv4", "address", "broadcast", "netmask"),
    ),
    (
        r"^inet6 addr: ([0-9a-z:]+)\/([0-9]+) Scope:Global",
        ("ipv6", "address", "mask_bits"),
    ),
]

nettools_2_regexes = [
    (
        r"^inet ([0-9\.]+)\s+netmask ([0-9\.fx]+)(?:\s+broadcast ([0-9\.]+))?$",
        ("ipv4", "address", "netmask", "broadcast"),
    ),
    (
        r"^inet6 ([0-9a-z:]+)\s+prefixlen ([0-9]+)",
        ("ipv6", "address", "mask_bits"),
    ),
]

iproute2_regexes = [
    (
        r"^inet ([0-9\.]+)\/([0-9]{1,2})(?:\s+brd ([0-9\.]+))?",
        ("ipv4", "address", "mask_bits", "broadcast"),
    ),
    (
        r"^inet6 ([0-9a-z:]+)\/([0-9]{1,3})",
        ("ipv6", "address", "mask_bits"),
    ),
]


def _parse_regexes(regexes, lines):
    data = {
        "ipv4": {},
        "ipv6": {},
    }

    for line in lines:
        for regex, groups in regexes:
            matches = re.match(regex, line)
            if matches:
                ip_data = {}

                for i, group in enumerate(groups[1:]):
                    ip_data[group] = matches.group(i + 1)

                if "mask_bits" in ip_data:
                    ip_data["mask_bits"] = int(ip_data["mask_bits"])

                target_group = data[groups[0]]
                if target_group.get("address"):
                    target_group.setdefault("additional_ips", []).append(ip_data)
                else:
                    target_group.update(ip_data)

                break

    return data


class NetworkDevices(FactBase):
    """
    Gets & returns a dict of network devices. See the ``ipv4_addresses`` and
    ``ipv6_addresses`` facts for easier-to-use shortcuts to get device addresses.

    .. code:: python

        {
            "eth0": {
                "ipv4": {
                    "address": "127.0.0.1",
                    "broadcast": "127.0.0.13",
                    # Only one of these will exist:
                    "netmask": "255.255.255.255",
                    "mask_bits": 32,
                },
                "ipv6": {
                    "address": "fe80::a00:27ff:fec3:36f0",
                    "mask_bits": 64,
                    "additional_ips": [{
                        "address": "fe80::",
                        "mask_bits": 128,
                    }],
                }
            },
        }
    """

    command = "ip addr show 2> /dev/null || ifconfig"
    default = dict

    # Definition of valid interface names for Linux:
    # https://git.kernel.org/pub/scm/linux/kernel/git/stable/linux.git/tree/net/core/dev.c?h=v5.1.3#n1020
    _start_regexes = [
        (
            r"^([^/: \s]+)\s+Link encap:",
            lambda lines: _parse_regexes(nettools_1_regexes, lines),
        ),
        (
            r"^([^/: \s]+): flags=",
            lambda lines: _parse_regexes(nettools_2_regexes, lines),
        ),
        (
            r"^[0-9]+: ([^/: \s]+): ",
            lambda lines: _parse_regexes(iproute2_regexes, lines),
        ),
    ]

    def process(self, output):
        devices = {}

        # Store current matches (start lines), the handler and any lines
        matches = None
        handler = None
        line_buffer = []

        for line in output:
            line = line.strip()

            matched = False

            # Look for start lines
            for regex, new_handler in self._start_regexes:
                new_matches = re.match(regex, line)

                # If we find a start line
                if new_matches:
                    matched = True

                    # Assign any current matches with current handler, reset buffer
                    if matches:
                        devices[matches.group(1)] = handler(line_buffer)
                        line_buffer = []

                    # Set new matches/handler
                    matches = new_matches
                    handler = new_handler
                    break

            if not matched:
                line_buffer.append(line)

        # Handle any left over matches
        if matches:
            devices[matches.group(1)] = handler(line_buffer)

        return devices


class Ipv4Addrs(ShortFactBase):
    """
    Gets & returns a dictionary of network interface -> list of IPv4 addresses.

    .. code:: python

        {
            "eth0": ["127.0.0.1"],
        }

    .. note::
        Network interfaces with no IPv4 will not be part of the dictionary.
    """

    fact = NetworkDevices
    ip_type = "ipv4"

    def process_data(self, data):
        host_to_ips = {}

        for interface, details in data.items():
            ips = []

            ip_details = details.get(self.ip_type)
            if not ip_details:
                continue

            ips.append(ip_details["address"])
            if "additional_ips" in ip_details:
                ips.extend([ip["address"] for ip in ip_details["additional_ips"]])

            host_to_ips[interface] = ips

        return host_to_ips


class Ipv6Addrs(Ipv4Addrs):
    """
    Gets & returns a dictionary of network interface -> list of IPv6 addresses.

    .. code:: python

        {
            "eth0": ["fe80::a00:27ff::2"],
        }

    .. note::
        Network interfaces with no IPv6 will not be part of the dictionary.
    """

    ip_type = "ipv6"


# TODO: remove these in v3
# Legacy versions of the above that only support one IP per interface
#


class Ipv4Addresses(ShortFactBase):
    """
    Gets & returns a dictionary of network interface -> IPv4 address.

    .. code:: python

        {
            "eth0": "127.0.0.1",
        }

    .. warning::
        This fact is deprecated, please use the ``hardware.Ipv4Addrs`` fact.

    .. note::
        Network interfaces with no IPv4 will not be part of the dictionary.
    """

    fact = NetworkDevices
    ip_type = "ipv4"

    def process_data(self, data):
        addresses = {}

        for interface, details in data.items():
            ip_details = details.get(self.ip_type)
            if not ip_details:
                continue  # pramga: no cover

            addresses[interface] = ip_details["address"]

        return addresses


class Ipv6Addresses(Ipv4Addresses):
    """
    Gets & returns a dictionary of network interface -> IPv6 address.

    .. code:: python

        {
            "eth0": "fe80::a00:27ff::2",
        }

    .. warning::
        This fact is deprecated, please use the ``hardware.Ipv6Addrs`` fact.

    .. note::
        Network interfaces with no IPv6 will not be part of the dictionary.
    """

    ip_type = "ipv6"



from os import path

import click

import pyinfra
from pyinfra import config, host, logger, state
from pyinfra.api.exceptions import PyinfraError
from pyinfra.api.util import get_file_path
from pyinfra.connectors.util import run_local_process, split_combined_output
from pyinfra.context import ctx_state


def include(filename: str):
    """
    Executes a local python file within the ``pyinfra.state.cwd``
    directory.
    """

    if not pyinfra.is_cli:
        raise PyinfraError("local.include is only available in CLI mode.")

    filename = get_file_path(state, filename)

    logger.debug("Including local file: %s", filename)

    config_state = config.get_current_state()

    try:
        # Fixes a circular import because `pyinfra.local` is really a CLI
        # only thing (so should be `pyinfra_cli.local`). It is kept here
        # to maintain backwards compatibility and the nicer public import
        # (ideally users never need to import from `pyinfra_cli`).

        from pyinfra_cli.util import exec_file

        with host.deploy(path.relpath(filename, state.cwd), None, None, in_deploy=False):
            exec_file(filename)

        # One potential solution to the above is to add local as an actual
        # module, ie `pyinfra.operations.local`.

    except Exception as e:
        raise PyinfraError(
            "Could not include local file: {0}:\n{1}".format(filename, e),
        )

    finally:
        config.set_current_state(config_state)


def shell(
    commands,
    splitlines: bool = False,
    ignore_errors: bool = False,
    print_output: bool = False,
    print_input: bool = False,
):
    """
    Subprocess based implementation of pyinfra/api/ssh.py's ``run_shell_command``.

    Args:
        commands (string, list): command or list of commands to execute
        spltlines (bool): optionally have the output split by lines
        ignore_errors (bool): ignore errors when executing these commands
    """

    if isinstance(commands, str):
        commands = [commands]

    all_stdout = []

    # Checking for state context being set means this function works outside a deploy
    # e.g.: the vagrant connector.
    if ctx_state.isset():
        print_output = state.print_output
        print_input = state.print_input

    for command in commands:
        print_prefix = "localhost: "

        if print_input:
            click.echo("{0}>>> {1}".format(print_prefix, command), err=True)

        return_code, combined_output = run_local_process(
            command,
            print_output=print_output,
            print_prefix=print_prefix,
        )
        stdout, stderr = split_combined_output(combined_output)

        if return_code > 0 and not ignore_errors:
            raise PyinfraError(
                "Local command failed: {0}\n{1}".format(command, stderr),
            )

        all_stdout.extend(stdout)

    if not splitlines:
        return "\n".join(all_stdout)

    return all_stdout



"""
.. warning::
    This connector is in alpha and may change in future releases.

Some Windows facts and Windows operations work but this is to be considered
experimental. For now, only ``winrm_username`` and ``winrm_password`` is
being used. There are other methods for authentication, but they have not yet
been added/experimented with.

The ``@winrm`` connector can be used to communicate with Windows instances that have WinRM enabled.

Examples using ``@winrm``:

.. code:: python

    # Get the windows_home fact
    pyinfra @winrm/192.168.3.232 --winrm-username vagrant \\
        --winrm-password vagrant --winrm-port 5985 -vv --debug fact windows_home

    # Create a directory
    pyinfra @winrm/192.168.3.232 --winrm-username vagrant \\
        --winrm-password vagrant --winrm-port 5985 windows_files.windows_directory 'c:\temp'

    # Run a powershell command ('ps' is the default shell-executable for the winrm connector)
    pyinfra @winrm/192.168.3.232 --winrm-username vagrant \\
        --winrm-password vagrant --winrm-port 5985 exec -- write-host hello

    # Run a command using the command prompt:
    pyinfra @winrm/192.168.3.232 --winrm-username vagrant \\
        --winrm-password vagrant --winrm-port 5985 --shell-executable cmd exec -- date /T

    # Run a command using the winrm ntlm transport
    pyinfra @winrm/192.168.3.232 --winrm-username vagrant \\
        --winrm-password vagrant --winrm-port 5985 --winrm-transport ntlm exec -- hostname
"""

import base64
import ntpath

import click

from pyinfra import logger
from pyinfra.api.connectors import BaseConnectorMeta
from pyinfra.api.exceptions import ConnectError, PyinfraError
from pyinfra.api.util import get_file_io, memoize, sha1_hash

from .pyinfrawinrmsession import PyinfraWinrmSession
from .util import make_win_command


class Meta(BaseConnectorMeta):
    handles_execution = True
    keys_prefix = "winrm"

    class DataKeys:
        hostname = "WinRM hostname to connect to"
        port = "WinRM port to connect to"
        user = "WinRM username"
        password = "WinRM password"
        transport = "WinRM transport (default: ``plaintext``)"
        read_timeout_sec = "Read timeout in seconds (default: ``30``)"
        operation_timeout_sec = "Operation timeout in seconds (default: ``20``)"


DATA_KEYS = Meta.keys()


def _raise_connect_error(host, message, data):
    message = "{0} ({1})".format(message, data)
    raise ConnectError(message)


@memoize
def show_warning():
    logger.warning("The @winrm connector is alpha!")


def _make_winrm_kwargs(state, host):
    kwargs = {}

    for key, value in (
        ("username", host.data.get(DATA_KEYS.user)),
        ("password", host.data.get(DATA_KEYS.password)),
        ("winrm_port", int(host.data.get(DATA_KEYS.port, 0))),
        ("winrm_transport", host.data.get(DATA_KEYS.transport, "plaintext")),
        (
            "winrm_read_timeout_sec",
            host.data.get(DATA_KEYS.read_timeout_sec, 30),
        ),
        (
            "winrm_operation_timeout_sec",
            host.data.get(DATA_KEYS.operation_timeout_sec, 20),
        ),
    ):
        if value:
            kwargs[key] = value

    # FUTURE: add more auth
    # pywinrm supports: basic, certificate, ntlm, kerberos, plaintext, ssl, credssp
    # see https://github.com/diyan/pywinrm/blob/master/winrm/__init__.py#L12

    return kwargs


def make_names_data(hostname):

    show_warning()

    yield "@winrm/{0}".format(hostname), {"winrm_hostname": hostname}, []


def connect(state, host):
    """
    Connect to a single host. Returns the winrm Session if successful.
    """

    kwargs = _make_winrm_kwargs(state, host)
    logger.debug("Connecting to: %s (%s)", host.name, kwargs)

    # Hostname can be provided via winrm config (alias), data, or the hosts name
    hostname = kwargs.pop(
        "hostname",
        host.data.get(DATA_KEYS.hostname, host.name),
    )

    try:
        # Create new session
        host_and_port = "{}:{}".format(hostname, host.data.get(DATA_KEYS.port))
        logger.debug("host_and_port: %s", host_and_port)

        session = PyinfraWinrmSession(
            host_and_port,
            auth=(
                kwargs["username"],
                kwargs["password"],
            ),
            transport=kwargs["winrm_transport"],
            read_timeout_sec=kwargs["winrm_read_timeout_sec"],
            operation_timeout_sec=kwargs["winrm_operation_timeout_sec"],
        )

        return session

    # TODO: add exceptions here
    except Exception as e:
        auth_kwargs = {}

        for key, value in kwargs.items():
            if key in ("username", "password"):
                auth_kwargs[key] = value

        auth_args = ", ".join("{0}={1}".format(key, value) for key, value in auth_kwargs.items())
        logger.debug("%s", e)
        _raise_connect_error(host, "Authentication error", auth_args)


def run_shell_command(
    state,
    host,
    command,
    env=None,
    success_exit_codes=None,
    print_output=False,
    print_input=False,
    return_combined_output=False,
    shell_executable=None,
    **ignored_command_kwargs,
):
    """
    Execute a command on the specified host.

    Args:
        state (``pyinfra.api.State`` obj): state object for this command
        hostname (string): hostname of the target
        command (string): actual command to execute
        success_exit_codes (list): all values in the list that will return success
        print_output (boolean): print the output
        print_intput (boolean): print the input
        return_combined_output (boolean): combine the stdout and stderr lists
        shell_executable (string): shell to use - 'cmd'=cmd, 'ps'=powershell(default)
        env (dict): environment variables to set

    Returns:
        tuple: (exit_code, stdout, stderr)
        stdout and stderr are both lists of strings from each buffer.
    """

    command = make_win_command(command)

    logger.debug("Running command on %s: %s", host.name, command)

    if print_input:
        click.echo("{0}>>> {1}".format(host.print_prefix, command), err=True)

    # get rid of leading/trailing quote
    tmp_command = command.strip("'")

    if print_output:
        click.echo(
            "{0}>>> {1}".format(host.print_prefix, command),
            err=True,
        )

    if not shell_executable:
        shell_executable = "ps"
    logger.debug("shell_executable:%s", shell_executable)

    # we use our own subclassed session that allows for env setting from open_shell.
    if shell_executable in ["cmd"]:
        response = host.connection.run_cmd(tmp_command, env=env)
    else:
        response = host.connection.run_ps(tmp_command, env=env)

    return_code = response.status_code
    logger.debug("response:%s", response)

    std_out_str = response.std_out.decode("utf-8")
    std_err_str = response.std_err.decode("utf-8")

    # split on '\r\n' (windows newlines)
    std_out = std_out_str.split("\r\n")
    std_err = std_err_str.split("\r\n")

    logger.debug("std_out:%s", std_out)
    logger.debug("std_err:%s", std_err)

    if print_output:
        click.echo(
            "{0}>>> {1}".format(host.print_prefix, "\n".join(std_out)),
            err=True,
        )

    if success_exit_codes:
        status = return_code in success_exit_codes
    else:
        status = return_code == 0

    logger.debug("Command exit status: %s", status)

    if return_combined_output:
        std_out = [("stdout", line) for line in std_out]
        std_err = [("stderr", line) for line in std_err]
        return status, std_out + std_err

    return status, std_out, std_err


def get_file(
    state, host, remote_filename, filename_or_io, remote_temp_filename=None, **command_kwargs
):
    raise PyinfraError("Not implemented")


def _put_file(state, host, filename_or_io, remote_location, chunk_size=2048):
    # this should work fine on smallish files, but there will be perf issues
    # on larger files both due to the full read, the base64 encoding, and
    # the latency when sending chunks
    with get_file_io(filename_or_io) as file_io:
        data = file_io.read()
        for i in range(0, len(data), chunk_size):
            chunk = data[i : i + chunk_size]
            ps = (
                '$data = [System.Convert]::FromBase64String("{0}"); '
                '{1} -Value $data -Encoding byte -Path "{2}"'
            ).format(
                base64.b64encode(chunk).decode("utf-8"),
                "Set-Content" if i == 0 else "Add-Content",
                remote_location,
            )
            status, _stdout, stderr = run_shell_command(state, host, ps)
            if status is False:
                logger.error("File upload error: {0}".format("\n".join(stderr)))
                return False

    return True


def put_file(
    state,
    host,
    filename_or_io,
    remote_filename,
    print_output=False,
    print_input=False,
    remote_temp_filename=None,  # ignored
    **command_kwargs,
):
    """
    Upload file by chunking and sending base64 encoded via winrm
    """

    # TODO: fix this? Workaround for circular import
    from pyinfra.facts.windows_files import WindowsTempDir

    # Always use temp file here in case of failure
    temp_file = ntpath.join(
        host.get_fact(WindowsTempDir),
        "pyinfra-{0}".format(sha1_hash(remote_filename)),
    )

    if not _put_file(state, host, filename_or_io, temp_file):
        return False

    # Execute run_shell_command w/sudo and/or su_user
    command = "Move-Item -Path {0} -Destination {1} -Force".format(temp_file, remote_filename)
    status, _, stderr = run_shell_command(
        state, host, command, print_output=print_output, print_input=print_input, **command_kwargs
    )

    if status is False:
        logger.error("File upload error: {0}".format("\n".join(stderr)))
        return False

    if print_output:
        click.echo(
            "{0}file uploaded: {1}".format(host.print_prefix, remote_filename),
            err=True,
        )

    return True



from unittest import TestCase

from paramiko import SSHException

from pyinfra.api import Config, State
from pyinfra.api.connect import connect_all
from pyinfra.api.exceptions import NoGroupError, NoHostError, PyinfraError

from ..paramiko_util import PatchSSHTestCase
from ..util import make_inventory


class TestInventoryApi(TestCase):
    def test_inventory_creation(self):
        inventory = make_inventory()

        # Check length
        assert len(inventory.hosts) == 2

        # Get a host
        host = inventory.get_host("somehost")
        assert host.data.ssh_user == "vagrant"

        # Check our group data
        assert inventory.get_group_data("test_group") == {
            "group_data": "hello world",
        }

    def test_tuple_host_group_inventory_creation(self):
        inventory = make_inventory(
            hosts=[
                ("somehost", {"some_data": "hello"}),
            ],
            tuple_group=(
                [
                    ("somehost", {"another_data": "world"}),
                ],
                {
                    "tuple_group_data": "word",
                },
            ),
        )

        # Check host data
        host = inventory.get_host("somehost")
        assert host.data.some_data == "hello"
        assert host.data.another_data == "world"

        # Check group data
        assert host.data.tuple_group_data == "word"

    def test_host_and_group_errors(self):
        inventory = make_inventory()

        with self.assertRaises(NoHostError):
            inventory.get_host("i-dont-exist")

        with self.assertRaises(NoGroupError):
            inventory.get_group("i-dont-exist")


class TestStateApi(PatchSSHTestCase):
    def test_fail_percent(self):
        inventory = make_inventory(
            (
                "somehost",
                ("thinghost", {"ssh_hostname": SSHException}),
                "anotherhost",
            ),
        )
        state = State(inventory, Config(FAIL_PERCENT=1))

        # Ensure we would fail at this point
        with self.assertRaises(PyinfraError) as context:
            connect_all(state)

        assert context.exception.args[0] == "Over 1% of hosts failed (33%)"

        # Ensure the other two did connect
        assert len(state.active_hosts) == 2



def handle_service_control(
    host,
    name,
    statuses,
    formatter,
    running,
    restarted,
    reloaded,
    command,
    status_argument="status",
):
    status = statuses.get(name, None)

    # If we don't know the status, we need to check if it's up before starting
    # and/or restarting/reloading
    if status is None:
        yield (
            "if ({status_command}); then "
            "({stop_command}); ({restart_command}); ({reload_command}); "
            "else ({start_command}); fi"
        ).format(
            status_command=formatter.format(name, status_argument),
            start_command=(formatter.format(name, "start") if running is True else "true"),
            stop_command=(formatter.format(name, "stop") if running is False else "true"),
            restart_command=(formatter.format(name, "restart") if restarted else "true"),
            reload_command=(formatter.format(name, "reload") if reloaded else "true"),
        )
        statuses[name] = running

    else:
        # Need down but running
        if running is False:
            if status:
                yield formatter.format(name, "stop")
                statuses[name] = False
            else:
                host.noop("service {0} is stopped".format(name))

        # Need running but down
        if running is True:
            if not status:
                yield formatter.format(name, "start")
                statuses[name] = True
            else:
                host.noop("service {0} is running".format(name))

        # Only restart if the service is already running
        if restarted and status:
            yield formatter.format(name, "restart")

        # Only reload if the service is already reloaded
        if reloaded and status:
            yield formatter.format(name, "reload")

    # Always execute arbitrary commands as these may or may not rely on the service
    # being up or down
    if command:
        yield formatter.format(name, command)



# encoding: utf-8

from unittest import TestCase

from pyinfra.api import Config, State
from pyinfra.connectors.util import (
    make_unix_command,
    make_unix_command_for_host,
    split_combined_output,
)

from ..util import make_inventory


class TestConnectorUtil(TestCase):
    def test_split_combined_output_works(self):
        results = split_combined_output(
            [
                ("stdout", "stdout1"),
                ("stdout", "stdout2"),
                ("stderr", "stderr1"),
                ("stdout", "stdout3"),
            ],
        )

        assert results == (["stdout1", "stdout2", "stdout3"], ["stderr1"])

    def test_split_combined_output_raises(self):
        with self.assertRaises(ValueError):
            split_combined_output(["nope", ""])


class TestMakeUnixCommandConnectorUtil(TestCase):
    def test_command(self):
        command = make_unix_command("echo Šablony")
        assert command.get_raw_value() == "sh -c 'echo Šablony'"

    def test_doas_command(self):
        command = make_unix_command("uptime", doas=True)
        assert command.get_raw_value() == "doas -n sh -c uptime"

    def test_doas_user_command(self):
        command = make_unix_command("uptime", doas=True, doas_user="pyinfra")
        assert command.get_raw_value() == "doas -n -u pyinfra sh -c uptime"

    def test_sudo_command(self):
        command = make_unix_command("uptime", sudo=True)
        assert command.get_raw_value() == "sudo -H -n sh -c uptime"

    def test_sudo_multi_arg_command(self):
        command = make_unix_command("echo hi", sudo=True, preserve_sudo_env=True)
        assert command.get_raw_value() == "sudo -H -n -E sh -c 'echo hi'"

    def test_sudo_preserve_env_command(self):
        command = make_unix_command("uptime", sudo=True, preserve_sudo_env=True)
        assert command.get_raw_value() == "sudo -H -n -E sh -c uptime"

    def test_use_sudo_login_command(self):
        command = make_unix_command("uptime", sudo=True, use_sudo_login=True)
        assert command.get_raw_value() == "sudo -H -n -i sh -c uptime"

    def test_sudo_user_command(self):
        command = make_unix_command("uptime", sudo=True, sudo_user="pyinfra")
        assert command.get_raw_value() == "sudo -H -n -u pyinfra sh -c uptime"

    def test_su_command(self):
        command = make_unix_command("uptime", su_user="pyinfra")
        assert command.get_raw_value() == "su pyinfra -c 'sh -c uptime'"

    def test_su_multi_arg_command(self):
        command = make_unix_command("echo hi", su_user="pyinfra")
        assert command.get_raw_value() == "su pyinfra -c 'sh -c '\"'\"'echo hi'\"'\"''"

    def test_use_su_login_command(self):
        command = make_unix_command("uptime", su_user="pyinfra", use_su_login=True)
        assert command.get_raw_value() == "su -l pyinfra -c 'sh -c uptime'"

    def test_preserve_su_env_command(self):
        command = make_unix_command("uptime", su_user="pyinfra", preserve_su_env=True)
        assert command.get_raw_value() == "su -m pyinfra -c 'sh -c uptime'"

    def test_su_shell_command(self):
        command = make_unix_command("uptime", su_user="pyinfra", su_shell="bash")
        assert command.get_raw_value() == "su -s `which bash` pyinfra -c 'sh -c uptime'"

    def test_command_env(self):
        command = make_unix_command(
            "uptime",
            env={
                "key": "value",
                "anotherkey": "anothervalue",
            },
        )
        assert command.get_raw_value() in [
            'sh -c \'export "key=value" "anotherkey=anothervalue" && uptime\'',
            'sh -c \'export "anotherkey=anothervalue" "key=value" && uptime\'',
        ]

    def test_command_chdir(self):
        command = make_unix_command("uptime", chdir="/opt/somedir")
        assert command.get_raw_value() == "sh -c 'cd /opt/somedir && uptime'"

    def test_custom_shell_command(self):
        command = make_unix_command("uptime", shell_executable="bash")
        assert command.get_raw_value() == "bash -c uptime"

    def test_mixed_command(self):
        command = make_unix_command(
            "echo hi",
            chdir="/opt/somedir",
            env={"key": "value"},
            sudo=True,
            sudo_user="root",
            preserve_sudo_env=True,
            su_user="pyinfra",
            shell_executable="bash",
        )
        assert command.get_raw_value() == (
            "sudo -H -n -E -u root "  # sudo bit
            "su pyinfra -c "  # su bit
            "'bash -c '\"'\"'cd /opt/somedir && export \"key=value\" "  # shell and export bit
            "&& echo hi'\"'\"''"  # command bit
        )

    def test_command_exists_su_config_only(self):
        """
        This tests covers a bug that appeared when `make_unix_command` is called
        with `su_user=False` (default) but `SU_USER` set on the config object,
        resulting in an empty command output.
        """
        state = State(make_inventory(), Config(SU_USER=True))
        host = state.inventory.get_host("somehost")
        command = make_unix_command_for_host(state, host, "echo Šablony")
        assert command.get_raw_value() == "sh -c 'echo Šablony'"



import re
from datetime import datetime

from pyinfra.api import QuoteString, StringCommand


def unix_path_join(*parts):
    parts = list(parts)
    parts[0:-1] = [part.rstrip("/") for part in parts[0:-1]]
    return "/".join(parts)


def ensure_mode_int(mode):
    # Already an int (/None)?
    if isinstance(mode, int) or mode is None:
        return mode

    try:
        # Try making an int ('700' -> 700)
        return int(mode)

    except (TypeError, ValueError):
        pass

    # Return as-is (ie +x which we don't need to normalise, it always gets run)
    return mode


def get_timestamp():
    return datetime.now().strftime("%y%m%d%H%M")


def sed_replace(
    filename,
    line,
    replace,
    flags=None,
    backup=False,
    interpolate_variables=False,
):
    flags = "".join(flags) if flags else ""

    line = line.replace("/", r"\/")
    replace = str(replace)
    replace = replace.replace("/", r"\/")
    backup_extension = get_timestamp()

    if interpolate_variables:
        line = line.replace('"', '\\"')
        replace = replace.replace('"', '\\"')
        sed_script_formatter = '"s/{0}/{1}/{2}"'
    else:
        # Single quotes cannot contain other single quotes, even when escaped , so turn
        # each ' into '"'"' (end string, double quote the single quote, (re)start string)
        line = line.replace("'", "'\"'\"'")
        replace = replace.replace("'", "'\"'\"'")
        sed_script_formatter = "'s/{0}/{1}/{2}'"

    sed_script = sed_script_formatter.format(line, replace, flags)

    sed_command = StringCommand(
        "sed",
        "-i.{0}".format(backup_extension),
        sed_script,
        QuoteString(filename),
    )

    if not backup:  # if we're not backing up, remove the file *if* sed succeeds
        backup_filename = "{0}.{1}".format(filename, backup_extension)
        sed_command = StringCommand(sed_command, "&&", "rm", "-f", QuoteString(backup_filename))

    return sed_command


def chmod(target, mode, recursive=False):
    args = ["chmod"]
    if recursive:
        args.append("-R")

    args.append("{0}".format(mode))

    return StringCommand(" ".join(args), QuoteString(target))


def chown(target, user, group=None, recursive=False, dereference=True):
    command = "chown"
    user_group = None

    if user and group:
        user_group = "{0}:{1}".format(user, group)

    elif user:
        user_group = user

    elif group:
        command = "chgrp"
        user_group = group

    args = [command]
    if recursive:
        args.append("-R")

    if not dereference:
        args.append("-h")

    return StringCommand(" ".join(args), user_group, QuoteString(target))


def adjust_regex(line, escape_regex_characters):
    """
    Ensure the regex starts with '^' and ends with '$' and escape regex characters if requested
    """
    match_line = line

    if escape_regex_characters:
        match_line = re.sub(r"([\.\\\+\*\?\[\^\]\$\(\)\{\}\-])", r"\\\1", match_line)

    # Ensure we're matching a whole line, note: match may be a partial line so we
    # put any matches on either side.
    if not match_line.startswith("^"):
        match_line = "^.*{0}".format(match_line)
    if not match_line.endswith("$"):
        match_line = "{0}.*$".format(match_line)

    return match_line



import json
from unittest import TestCase
from unittest.mock import mock_open, patch

from pyinfra.api.exceptions import InventoryError
from pyinfra.connectors.vagrant import get_vagrant_options, make_names_data

FAKE_VAGRANT_OPTIONS = {
    "groups": {
        "ubuntu16": ["mygroup"],
    },
    "data": {
        "centos7": {
            "somedata": "somevalue",
        },
    },
}
FAKE_VAGRANT_OPTIONS_DATA = json.dumps(FAKE_VAGRANT_OPTIONS)


def fake_vagrant_shell(command, splitlines=None):
    if command == "vagrant status --machine-readable":
        return [
            "1574707743,ubuntu16,provider-name,virtualbox",
            "1574707743,ubuntu16,state,running",
            "1574707743,ubuntu18,state,not_created",
            "1574707743,centos7,state,running",
            "1574707743,centos6,state,running",
        ]
    elif command == "vagrant ssh-config ubuntu16":
        return [
            "ExtraKey logme!",
            "Host ubuntu16",
            "HostName 127.0.0.1",
            "User vagrant",
            "Port 2222",
            "PasswordAuthentication no",
            "IdentityFile path/to/key",
            "",
        ]
    elif command == "vagrant ssh-config centos7":
        return [
            "",
            "Host centos7",
            "HostName 127.0.0.1",
            "User vagrant",
            "Port 2200",
            "PasswordAuthentication no",
            "IdentityFile path/to/key",
        ]
    elif command == "vagrant ssh-config centos6":
        return [
            "Host centos6",
            "HostName 127.0.0.1",
        ]

    return []


@patch("pyinfra.connectors.vagrant.local.shell", fake_vagrant_shell)
class TestVagrantConnector(TestCase):
    def tearDown(self):
        get_vagrant_options.cache = {}

    @patch(
        "pyinfra.connectors.vagrant.open",
        mock_open(read_data=FAKE_VAGRANT_OPTIONS_DATA),
        create=True,
    )
    @patch("pyinfra.connectors.vagrant.path.exists", lambda path: True)
    def test_make_names_data_with_options(self):
        data = make_names_data()

        assert data == [
            (
                "@vagrant/ubuntu16",
                {
                    "ssh_port": "2222",
                    "ssh_user": "vagrant",
                    "ssh_hostname": "127.0.0.1",
                    "ssh_key": "path/to/key",
                },
                ["mygroup", "@vagrant"],
            ),
            (
                "@vagrant/centos7",
                {
                    "ssh_port": "2200",
                    "ssh_user": "vagrant",
                    "ssh_hostname": "127.0.0.1",
                    "ssh_key": "path/to/key",
                    "somedata": "somevalue",
                },
                ["@vagrant"],
            ),
            (
                "@vagrant/centos6",
                {
                    "ssh_hostname": "127.0.0.1",
                },
                ["@vagrant"],
            ),
        ]

    def test_make_names_data_with_limit(self):
        data = make_names_data(limit=("ubuntu16",))

        assert data == [
            (
                "@vagrant/ubuntu16",
                {
                    "ssh_port": "2222",
                    "ssh_user": "vagrant",
                    "ssh_hostname": "127.0.0.1",
                    "ssh_key": "path/to/key",
                },
                ["@vagrant"],
            ),
        ]

    def test_make_names_data_no_matches(self):
        with self.assertRaises(InventoryError):
            make_names_data(limit="nope")
